import math
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple

import attr
import numpy as np
from scipy import stats

from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.errors import WorldTooSmall
from avalon.datagen.world_creation.configs.building import BuildingConfig
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import IndoorTaskConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import AGENT_HEIGHT
from avalon.datagen.world_creation.constants import JUMPING_REQUIRED_HEIGHT
from avalon.datagen.world_creation.constants import MAX_EFFECTIVE_JUMP_DIST
from avalon.datagen.world_creation.constants import MAX_FALL_DISTANCE_TO_DIE
from avalon.datagen.world_creation.constants import MAX_FLAT_JUMP_DIST
from avalon.datagen.world_creation.constants import MAX_JUMP_HEIGHT_METERS
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.entities.entity import Entity
from avalon.datagen.world_creation.geometry import BuildingTile
from avalon.datagen.world_creation.geometry import local_to_global_coords
from avalon.datagen.world_creation.indoor.builders import DefaultHallwayBuilder
from avalon.datagen.world_creation.indoor.builders import HouseLikeRoomBuilder
from avalon.datagen.world_creation.indoor.builders import RectangleFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import WindowBuilder
from avalon.datagen.world_creation.indoor.building import Building
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.indoor.components import FloorChasm
from avalon.datagen.world_creation.indoor.constants import CEILING_THICKNESS
from avalon.datagen.world_creation.indoor.constants import DEFAULT_FLOOR_THICKNESS
from avalon.datagen.world_creation.indoor.constants import MIN_BUILDING_SIZE
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.task_generator import CANONICAL_BUILDING_LOCATION
from avalon.datagen.world_creation.indoor.task_generator import BuildingTaskGenerator
from avalon.datagen.world_creation.indoor.task_generator import IndoorTaskParams
from avalon.datagen.world_creation.indoor.task_generator import create_building_for_skill_scenario
from avalon.datagen.world_creation.indoor.task_generator import make_indoor_task_world
from avalon.datagen.world_creation.indoor.task_generator import rectangle_dimensions_within_radius
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.obstacles.height_path import HeightPath
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations

# smaller than this and there might not end up being much of a gap at all
# has to do with the resolution with which terrain is created
_MIN_DIST_FOR_JUMP = 1.0

# minimum height that requires the jump button to be pressed at all (less than this and you can just walk over it)
_MIN_HEIGHT_REQUIRING_JUMP = 0.7


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class JumpIndoorTaskConfig(IndoorTaskConfig):
    # Smallest and largest possible site radius for a building with this task to be placed on.
    min_site_radius: float = 6.5
    max_site_radius: float = 35.0
    # Standard deviation for the distribution deciding the site radius: higher means more variability at same difficulty
    site_radius_std_dev: float = 3.0
    # Total number of stories for the building. Note that this task is a single-story task, all other stories will
    # be purely aesthetic (e.g. to help with being viewed in the distance outdoors).
    story_count: int = 2


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class JumpOutdoorTaskConfig(TaskConfig):
    # the easier versions of the jump task allow for the agent to simply fall in the pit and jump out
    # these parameters control the min and max heights for that mode
    jumpable_mode_height_min: float = _MIN_HEIGHT_REQUIRING_JUMP
    jumpable_mode_height_max: float = JUMPING_REQUIRED_HEIGHT
    # the harder mode of the jump task requires that the agent pick the correct place to jump over a very specially
    # crafted gap that is exactly as wide as can be jumped over. These parameters control that jump distance
    jump_dist_easy: float = _MIN_DIST_FOR_JUMP
    jump_dist_hard: float = MAX_EFFECTIVE_JUMP_DIST
    # how wide the area is where the gap is the correct distance. Beyond this width, it can be much farther across
    jump_region_width_easy: float = 10.0
    jump_region_width_hard: float = 3.0
    jump_region_width_std_dev: float = 1.0
    # how much to flatten around the jump area. This is a bit unfortunate, as it makes it easy to tell where to jump.
    # howver, without this, jumping is too easy when the terrain is uneven, because you can just find a place where
    # one side of the chasm happens to be higher than the other
    jump_region_flatten_width_easy: float = 8.0
    jump_region_flatten_width_hard: float = 4.0
    jump_region_flatten_width_std_dev: float = 1.0
    # how far away the goal is
    goal_dist_min: float = 7.0
    goal_dist_max: float = 12.0
    # how likely climbing is to be allowed at difficulty 1.0  (at 0.0 it's always allowed)
    ending_climb_probability: float = 0.1


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class JumpTaskConfig(TaskConfig):
    # likelihood that the indoor variant of the task will be used
    indoor_probability: float = 0.2
    # the configs for each variant of the task
    indoor_config: JumpIndoorTaskConfig = JumpIndoorTaskConfig()
    outdoor_config: JumpOutdoorTaskConfig = JumpOutdoorTaskConfig()


def generate_jump_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: JumpTaskConfig = JumpTaskConfig(),
) -> None:
    is_indoor = rand.uniform() < task_config.indoor_probability
    if is_indoor:
        building, entities, spawn_location, target_location = create_building_for_skill_scenario(
            rand,
            difficulty,
            JumpTaskGenerator(task_config.indoor_config),
            position=CANONICAL_BUILDING_LOCATION,
            is_indoor_only=True,
        )
        world = make_indoor_task_world(
            building, entities, difficulty, spawn_location, target_location, rand, export_config
        )
    else:
        world, locations, difficulty = create_jump_obstacle(
            rand, difficulty, export_config, task_config=task_config.outdoor_config
        )
        world, locations = world.end_height_obstacles(
            locations, is_accessible_from_water=False, is_spawn_region_climbable=False
        )
        world = add_food_tree_for_simple_task(world, locations)
        world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)

    export_world(output_path, rand, world)


def create_jump_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    task_config: JumpOutdoorTaskConfig = JumpOutdoorTaskConfig(),
) -> Tuple[World, WorldLocations, float]:

    is_depth_jumpable, difficulty = select_boolean_difficulty(difficulty, rand)
    is_climbing_possible, difficulty = select_boolean_difficulty(
        difficulty, rand, initial_prob=1, final_prob=task_config.ending_climb_probability
    )
    desired_goal_dist = difficulty_variation(task_config.goal_dist_min, task_config.goal_dist_max, rand, difficulty)

    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_dist, desired_goal_dist / 10), rand, difficulty, export_config, constraint
    )

    # make it not impossibly hard to jump
    desired_jump_distance = scale_with_difficulty(difficulty, task_config.jump_dist_easy, task_config.jump_dist_hard)
    jump_distance, _warnings = world.get_critical_distance(locations, _MIN_DIST_FOR_JUMP, desired_jump_distance)
    jumpable_width = normal_distrib_range(
        task_config.jump_region_width_easy,
        task_config.jump_region_width_hard,
        task_config.jump_region_width_std_dev,
        rand,
        difficulty,
    )

    if jump_distance is None:
        raise WorldTooSmall(AvalonTask.JUMP, _MIN_DIST_FOR_JUMP, locations.get_2d_spawn_goal_distance())

    if is_depth_jumpable:
        depth = difficulty_variation(
            task_config.jumpable_mode_height_min, task_config.jumpable_mode_height_max, rand, difficulty
        )
    else:
        # we need deeper gaps for the compositional tasks--because they can be on more difficult terrain,
        # the gap needs to be quite deep to prevent you from glitching your way across
        depth = scale_with_difficulty(difficulty, JUMPING_REQUIRED_HEIGHT, MAX_FALL_DISTANCE_TO_DIE)

    ring_config = make_ring(
        rand,
        difficulty,
        world,
        locations,
        jump_distance,
        constraint=constraint,
        height=-depth,
        traversal_width=jumpable_width,
        is_inside_climbable=True,
        is_outside_climbable=is_climbing_possible,
        dual_solution=HeightSolution(
            paths=(
                HeightPath(
                    is_solution_flattened=True,
                    is_height_affected=False,
                    width=normal_distrib_range(
                        task_config.jump_region_flatten_width_easy,
                        task_config.jump_region_flatten_width_hard,
                        task_config.jump_region_flatten_width_std_dev,
                        rand,
                        difficulty,
                    ),
                ),
            ),
            solution_point_brink_distance=1.0,
        ),
        extra_safety_radius=0.5,
    )
    world = world.add_height_obstacle(rand, ring_config, locations.island)

    return world, locations, difficulty


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class JumpTaskGenerator(BuildingTaskGenerator):
    config: JumpIndoorTaskConfig = JumpIndoorTaskConfig()

    def get_site_radius(self, rand: np.random.Generator, difficulty: float) -> float:
        return normal_distrib_range(
            self.config.min_site_radius, self.config.max_site_radius, self.config.site_radius_std_dev, rand, difficulty
        )

    def get_building_config(
        self,
        rand: np.random.Generator,
        difficulty: float,
        radius: float,
        allowed_auxiliary_tasks: Tuple[BuildingTask, ...] = tuple(),
        aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig(),
    ) -> BuildingConfig:
        width, length = rectangle_dimensions_within_radius(radius)
        width = max(width // 2, MIN_BUILDING_SIZE)
        room_height = ((MAX_JUMP_HEIGHT_METERS + AGENT_HEIGHT) * 1.1) + CEILING_THICKNESS + DEFAULT_FLOOR_THICKNESS

        return BuildingConfig(
            width=width,
            length=length,
            story_count=self.config.story_count,
            footprint_builder=RectangleFootprintBuilder(),
            room_builder=HouseLikeRoomBuilder(max_rooms=1, room_height=room_height),
            hallway_builder=DefaultHallwayBuilder(proportion_additional_edges=1),
            story_linker=None,
            window_builder=WindowBuilder(),
            aesthetics=aesthetics,
        )

    def get_principal_obstacle_params(self, rand: np.random.Generator, difficulty: float, building: Building) -> Tuple:
        spawn_story = building.stories[0]
        spawn_room = spawn_story.rooms[0]
        # Note: crossing a 3-wide chasm is possible, but very difficult, so not included below
        max_chasm_width = math.floor(MAX_FLAT_JUMP_DIST)

        # We need one tile for launch and two for landing: they must be separate to enable climbing only on one side,
        # and landing needs two, so you can't stretch arm out through to the next climbable surface and escape the chasm
        launch_platform_width = 1
        landing_platform_width = 2
        platform_width = launch_platform_width + landing_platform_width
        max_obstacle_width = max_chasm_width + platform_width
        max_obstacle_count = (spawn_room.length - 2) // (max_obstacle_width)
        if max_obstacle_count == 0:
            raise ImpossibleWorldError("Building too small for jump task")

        free_site_indices = [i * max_obstacle_width for i in range(max_obstacle_count)]
        obstacle_count = 1 + round(normal_distrib_range(0, max_obstacle_count - 1, 0.1, rand, difficulty))
        site_indices = rand.choice(free_site_indices, size=obstacle_count)
        chasms = []
        for site_index in site_indices:
            # Note on indices: site index is the start of the "obstacle", which includes both launching and landing
            # platforms and the chasm. The launch platform is index 0, chasm is 1-2 and landing platform is 3
            # These numbers will differ if we change the max jumping distance / chasm width
            chasm_thickness = round(1 + normal_distrib_range(0, max_chasm_width - 1, 0.1, rand, difficulty))

            # Chasm must be deep enough to avoid winning by jumping + grabbing
            chasm_depth = (MAX_JUMP_HEIGHT_METERS * 2) - DEFAULT_FLOOR_THICKNESS
            depth_map = np.full_like(spawn_room.floor_heightmap, np.nan, dtype=np.float32)
            chasm_start_index = site_index + launch_platform_width
            chasm_end_index = chasm_start_index + chasm_thickness
            depth_map[chasm_start_index:chasm_end_index, :] = chasm_depth
            climbable_mask = np.zeros_like(spawn_room.floor_heightmap, dtype=np.bool_)
            climbable_mask[site_index, :] = True
            obstacle_mask = np.zeros_like(spawn_room.floor_heightmap, dtype=np.bool_)
            obstacle_mask[site_index : chasm_end_index + landing_platform_width, :] = True
            chasms.append(FloorChasm(spawn_story.num, spawn_room.id, obstacle_mask, climbable_mask, depth_map))

        spawn_tile = rand.choice([BuildingTile(i, 0) for i in range(spawn_room.width)])  # type: ignore[arg-type]
        target_tile = rand.choice([BuildingTile(i, spawn_room.length - 2) for i in range(1, spawn_room.width - 1)])  # type: ignore[arg-type]
        entrance_azimuth = Azimuth.NORTH
        return spawn_tile, entrance_azimuth, target_tile, chasms

    def add_principal_obstacles(
        self, rand: np.random.Generator, building: Building, obstacle_params: Tuple
    ) -> IndoorTaskParams:
        spawn_tile, entrance_azimuth, target_tile, chasms = obstacle_params
        for chasm in chasms:
            chasm.apply(building.stories)

        first_chasm: FloorChasm = chasms[0]
        spawn_story = building.stories[first_chasm.story_id]
        spawn_room = spawn_story.rooms[first_chasm.room_id]
        spawn_location_in_room = self._position_from_tile(spawn_room, spawn_tile, AGENT_HEIGHT / 2)
        # Do not use FOOD_HOVER_DIST below as the food may fall into a nearby chasm if it is spawned close to a wall
        target_location_in_room = self._position_from_tile(spawn_room, target_tile, 0)

        room_offset = np.array([spawn_room.position.x, 0, spawn_room.position.z])
        spawn_location_in_building = local_to_global_coords(np.array(spawn_location_in_room), room_offset)
        target_location_in_building = local_to_global_coords(np.array(target_location_in_room), room_offset)

        extra_items: List[Entity] = []
        entrance_sites = ((spawn_story.num, spawn_room.id, tuple([entrance_azimuth])),)
        return building, extra_items, spawn_location_in_building, target_location_in_building, entrance_sites



"""A little utility library to make for easy logging to wandb."""
import math
import queue
import secrets
import time
from types import ModuleType
from typing import Optional
from typing import Union

import numpy as np
import torch
import wandb
from loguru import logger
from numpy import typing as npt
from torch import Tensor
from wandb.sdk.lib import RunDisabled
from wandb.sdk.wandb_run import Run

last_time: Optional[float] = None
last_step = 0
LOG_BACKOFF_POINT = 5000
LOG_BACKOFF_FACTOR = 20

# backoff does apply to scalars and media
SCALAR_FREQ = 5
MEDIA_FREQ = 100
# backoff does not apply to hists
HIST_FREQ = 100

# TODO: instead of every call including the step (and thus having to know the step in all code),
# we should just update the step once per iteration with a special call.
# Why didn't I do it that way in the first place? Does that limit any capabilities?


def effective_freq(step: int, freq: int) -> int:
    # Simple logging backoff logic.
    if step > LOG_BACKOFF_POINT and freq != 1:
        freq *= LOG_BACKOFF_FACTOR
    return freq


def check_log_interval(step: int, freq: int, backoff: bool = True) -> bool:
    if backoff:
        freq = effective_freq(step, freq)
    return step % freq == 0


def log_from_queue(wandb_queue, prefix: str = "") -> None:  # type: ignore
    """Allows logging from another process."""
    try:
        mode, key, value = wandb_queue.get_nowait()
        key = f"{prefix}{key}"
        if mode == "scalar":
            log_scalar(key, value, freq=1)
        elif mode == "histogram":
            log_histogram(key, value, mean_freq=1, hist_freq=1)
        elif mode == "video":
            log_video(key, value, freq=1)
        elif mode == "image":
            log_image(key, value, freq=1)
        else:
            logger.warning(f"unrecognized mode in log_from_pipe: {mode}, {key}")
    except queue.Empty:
        pass


def log_scalar(
    tag: str,
    value: Union[int, float, Tensor],
    step: Optional[int] = None,
    freq: Optional[int] = None,
    wandb_run: Union[ModuleType, Run, RunDisabled] = wandb,
) -> None:
    global last_step
    freq = SCALAR_FREQ if freq is None else freq
    if step is None:
        step = last_step
    else:
        last_step = step
    if not check_log_interval(step, freq):
        return
    if type(value) == torch.Tensor:
        value = value.cpu().detach()
    wandb_run.log({tag: value}, step=step)


def log_histogram(
    tag: str,
    value: Union[list, npt.NDArray, Tensor],
    step: Optional[int] = None,
    mean_freq: Optional[int] = None,
    hist_freq: Optional[int] = None,
    log_mean: bool = True,
) -> None:
    global last_step
    mean_freq = SCALAR_FREQ if mean_freq is None else mean_freq
    hist_freq = HIST_FREQ if hist_freq is None else hist_freq

    if step is None:
        step = last_step
    else:
        last_step = step

    # Allow backoff to apply to the scalars
    if log_mean:
        try:
            # works on torch and tf
            mean = value.mean()  # type: ignore
        except AttributeError:
            # hopefully works on most everything else
            mean = np.mean(value)
        log_scalar(tag + "_mean", mean, step, mean_freq)
    # Backoff doesn't work well with hists, messes up the viz
    if not check_log_interval(step, hist_freq, backoff=False):
        return
    if type(value) == torch.Tensor:
        value = value.cpu().detach()
    wandb.log({tag + "_hist": wandb.Histogram(value)}, step=step)  # type: ignore


def log_video(
    tag: str,
    batch: Tensor,
    step: Optional[int] = None,
    freq: Optional[int] = None,
    normalize: bool = False,
    num_images_per_row: int = 4,
    video_format: str = "mp4",  # mp4, webm. webm preferred except it was giving some weird bug
) -> None:
    global last_step
    freq = MEDIA_FREQ if freq is None else freq
    if step is None:
        step = last_step
    else:
        last_step = step
    # Expects b, t, c, h, w
    # or t, c, h, w
    if not check_log_interval(step, freq):
        return

    if normalize:
        min_v = torch.min(batch)
        range_v = torch.max(batch) - min_v
        if range_v > 0:
            batch = (batch - min_v) / range_v
        else:
            batch = torch.zeros(batch.size())

    # batch = preprocess_batch(batch).permute(0, 2, 1, 3, 4) + .5
    if len(batch.shape) == 5:
        frames = make_video_grid(batch, num_images_per_row=num_images_per_row, pad_value=1)
    else:
        assert len(batch.shape) == 4
        frames = batch

    # This should be in range 0-1
    if type(frames) == torch.Tensor:
        frames = frames.detach()
        if torch.is_floating_point(frames):
            frames = (frames * 255).clamp(0, 255).to(torch.uint8)
        frames = frames.cpu()
    elif isinstance(frames, np.ndarray):
        if np.issubdtype(frames.dtype, np.floating):
            frames = (frames * 255).clip(0, 255).astype(np.uint8)
    else:
        assert False

    # TODO: make this work for non-tensor
    # Wandb doesn't like 1-channel greyscale videos
    shape = list(frames.shape)
    if shape[1] == 1:
        shape[1] = 3
        frames = frames.expand(shape)
    wandb.log({tag: wandb.Video(frames, fps=4, format=video_format)}, step=step)  # type: ignore


def log_image(tag: str, value: Tensor, step: Optional[int] = None, freq: Optional[int] = None) -> None:
    global last_step
    if step is None:
        step = last_step
    else:
        last_step = step
    freq = MEDIA_FREQ if freq is None else freq
    if not check_log_interval(step, freq):
        return
    if type(value) == torch.Tensor:
        value = value.cpu().detach()
    wandb.log({tag: wandb.Image(value)}, step=step)


def log_table(tag: str, value: Tensor, step: int, freq: Optional[int] = None) -> None:
    freq = MEDIA_FREQ if freq is None else freq
    # value should be a 1d tensor, in this current implementation. can add more columns in the future.
    if not check_log_interval(step, freq):
        return
    if type(value) == torch.Tensor:
        value = value.cpu().detach()
    columns = ["test"]
    rows = [[x] for x in value]
    table = wandb.Table(data=rows, columns=columns)
    wandb.log({tag: table}, step=step)


def log_iteration_time(batch_size: int, step: int, freq: Optional[int] = None) -> None:
    """Batch size should be the number of samples processed per step of i."""
    freq = SCALAR_FREQ if freq is None else freq
    global last_time
    global last_step
    if not check_log_interval(step, freq):
        return

    if last_time is None:
        last_time = time.time()
        last_step = step
    else:
        if step == last_step:
            return
        dt = (time.time() - last_time) / (step - last_step)
        last_time = time.time()
        last_step = step
        log_scalar("timings/iterations-per-sec", 1 / dt, step, freq=1)
        log_scalar("timings/samples-per-sec", batch_size / dt, step, freq=1)


def watch(model: torch.nn.Module, freq: Optional[int] = None) -> None:
    freq = HIST_FREQ if freq is None else freq
    wandb.watch(model, log="all", log_freq=freq, log_graph=False)


def download_file(run_id: str, project_name: str, filename: Optional[str] = None) -> str:
    api = wandb.Api()
    run = api.run(f"sourceress/{project_name}/{run_id}")
    # We save to a random directory to avoid contention issues if there's
    # potentially multiple processes downloading the same file at the same time.
    # TODO: clean these files up
    path = run.file(filename).download(replace=True, root=f"./data/{secrets.token_hex(10)}").name
    assert isinstance(path, str)
    return path


def make_video_grid(
    tensor: Tensor,
    num_images_per_row: int = 10,
    padding: int = 2,
    pad_value: int = 0,
) -> Tensor:
    """
    This is a repurposed implementation of `make_grid` from torchvision to work with videos.

    Expects shape (batch_size, timesteps, channels, h, w).
    """
    n_maps, sequence_length, num_channels, height, width = tensor.size()
    x_maps = min(num_images_per_row, n_maps)
    y_maps = int(math.ceil(float(n_maps) / x_maps))
    height, width = int(height + padding), int(width + padding)
    grid = tensor.new_full(
        (sequence_length, num_channels, height * y_maps + padding, width * x_maps + padding), pad_value
    )
    k = 0
    for y in range(y_maps):
        for x in range(x_maps):
            if k >= n_maps:
                break
            grid.narrow(2, y * height + padding, height - padding).narrow(
                3, x * width + padding, width - padding
            ).copy_(tensor[k])
            k += 1
    return grid



import hashlib
import json
import os
import shutil
import sys
import time
import traceback
from multiprocessing import Pool
from pathlib import Path
from typing import List
from typing import Sequence
from typing import TypedDict

import requests
from loguru import logger

from avalon.common.log_utils import configure_parent_logging
from avalon.contrib.utils import run_local_command
from avalon.datagen.generate_worlds import GeneratedWorld
from avalon.datagen.godot_generated_types import AvalonSimSpec

EDITOR_SETTINGS = """
[gd_resource type="EditorSettings" format=2]
[resource]
export/android/android_sdk_path = "/opt/oculus/android-sdk"
export/android/shutdown_adb_on_exit = false
"""


def get_apk_filename(participant_id: str) -> str:
    return f"avalon__{participant_id}.apk"


class APKGenerationResult(TypedDict):
    participant_id: str
    apk_file: Path
    apk_hash: str


def generate_apk(
    godot_path: Path,
    worlds_path: Path,
    tmp_path: Path,
    output_path: Path,
    apk_script: str,
    participant_id: str,
    apk_version: str,
    include_worlds: bool = True,
    is_output_traced: bool = False,
):
    output_path.mkdir(parents=True, exist_ok=True)

    # create a dir in which we'll copy all the files in and begin doing the work
    participant_working_path = tmp_path / participant_id
    if participant_working_path.exists():
        shutil.rmtree(participant_working_path)
    participant_working_path.mkdir(parents=True)

    assert godot_path.exists(), f"Godot path not found: {godot_path}"

    participant_godot_path = participant_working_path / "godot"
    # copy godot folder in
    shutil.copytree(godot_path, participant_godot_path)

    (participant_godot_path / "export_presets.template.cfg").unlink()

    # update the user id and app version in the config file
    raw_config_path = participant_godot_path / "android" / "config.json"
    config = AvalonSimSpec.from_dict(json.load(open(raw_config_path, "r")))
    with config.mutable_clone() as new_config:
        new_config.recording_options.user_id = participant_id
        new_config.recording_options.apk_version = apk_version
    serialized_config = new_config.to_dict()
    json.dump(serialized_config, open(raw_config_path, "w"))

    pck_name = get_apk_filename(participant_id).replace(".apk", ".pck")
    apk_name = get_apk_filename(participant_id)
    pck_temp = participant_godot_path / pck_name
    apk_temp = participant_godot_path / apk_name
    apk_file = output_path / apk_name

    self_contained_godot_bin = participant_working_path / "godot_bin"
    (participant_working_path / "._sc_").touch()
    (participant_working_path / "editor_data").mkdir()
    (participant_working_path / "editor_data" / "editor_settings-3.tres").write_text(EDITOR_SETTINGS)

    pre_bake_command = f"cd {participant_godot_path} && {apk_script}"
    pack_command = f"cd {participant_godot_path} && {self_contained_godot_bin} --verbose --export-pack  Android {pck_temp} 2>&1 | tee ../build.pck.log"
    bake_command = f"cd {participant_godot_path} && {self_contained_godot_bin} --verbose --export-debug Android {apk_temp} 2>&1 | tee ../build.apk.log"

    # make sure build dependencies are installed, clear editor config
    run_local_command(pre_bake_command, trace_output=is_output_traced)

    # Actually make the self-contained bin available (now that pre-bake script has downloaded it)
    # We must first ensure there is a copy of it on the tmp path as hard links cannot be done across file systems
    tmp_godot_path = tmp_path / "godot"
    if not tmp_godot_path.exists():
        shutil.copyfile("/opt/oculus/godot/godot", tmp_godot_path)
    os.link(tmp_godot_path, self_contained_godot_bin)
    os.chmod(tmp_godot_path, 755)  # for some reason, os.link drops the executable permission

    # bake twice, the first build just fixes dependencies/imports
    run_local_command(pack_command, trace_output=is_output_traced)
    if include_worlds:
        assert worlds_path.exists(), f"World path not found: {worlds_path}"
        # link in worlds
        os.symlink(worlds_path, participant_godot_path / "worlds")
    run_local_command(bake_command, trace_output=is_output_traced)

    # move apk to output path
    apk_temp.rename(apk_file)
    pck_temp.unlink()

    # clean up and delete after
    # shutil.rmtree(participant_working_path)

    list_command = f"unzip -v {apk_file} -x 'META-INF/*' -x 'assets/android/config.json' | cut -c 48- | tail -n+2"
    apk_contents = run_local_command(list_command, trace_output=False).stdout
    (participant_working_path / "build.apk.lst").write_bytes(apk_contents)

    return {
        "participant_id": participant_id,
        "apk_file": apk_file,
        "apk_hash": hashlib.sha1(apk_contents).hexdigest(),
    }


def generate_apks(
    godot_path: Path,
    worlds_path: Path,
    tmp_path: Path,
    output_path: Path,
    apk_script: str,
    participant_ids: List[str],
    apk_version: str,
    include_worlds: bool = True,
    is_output_traced: bool = False,
):
    errors = []

    def on_done(result: APKGenerationResult) -> None:
        pass

    def on_error(error: BaseException) -> None:
        traceback.print_exception(type(error), error, sys.exc_info()[2])
        errors.append(error)

    start_time = time.time()

    if not output_path.exists():
        output_path.mkdir(parents=True)

    with Pool(processes=len(participant_ids), initializer=configure_parent_logging) as worker_pool:
        requests = {}
        for participant_id in participant_ids:
            logger.info(f"adding {participant_id} to requests")
            request = worker_pool.apply_async(
                generate_apk,
                kwds={
                    "godot_path": godot_path,
                    "worlds_path": worlds_path,
                    "tmp_path": tmp_path,
                    "output_path": output_path,
                    "apk_script": apk_script,
                    "participant_id": participant_id,
                    "apk_version": apk_version,
                    "include_worlds": include_worlds,
                    "is_output_traced": is_output_traced,
                },
                callback=on_done,
                error_callback=on_error,
            )
            requests[participant_id] = request

        results = []
        for participant_id, request in requests.items():
            request.wait()
            if request.successful():
                results.append(request.get())
            else:
                results.append(None)
                logger.error(f"ERROR: {tmp_path / participant_id}/build.apk.log")

    logger.info(f"finished in {time.time() - start_time} seconds")

    return results


AVALON_SERVER_URL = "http://avalon.int8.ai:64080"


def add_worlds_to_server(worlds: Sequence[GeneratedWorld]) -> None:
    world_ids = [world.world_id for world in worlds]

    for world_id in world_ids:
        r = requests.get(f"{AVALON_SERVER_URL}/make_world/{world_id}/")
        assert r.status_code == 200

    r = requests.get(f"{AVALON_SERVER_URL}/get_state/")
    assert r.status_code == 200
    state = r.json()["state"]
    assert set(world_ids).issubset(set(state["user_ids_by_world_id"].keys()))


def add_apk_version_to_server(apk_version: str) -> None:
    r = requests.get(f"{AVALON_SERVER_URL}/add_apk_version/{apk_version}/")
    assert r.status_code == 200


def upload_apk_to_server(path: Path):
    files = {"file": path.open("rb")}
    r = requests.post(f"{AVALON_SERVER_URL}/upload_apk/", files=files)
    assert r.status_code == 200
    return r.json()



import math
from itertools import product
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import cast

import attr
import numpy as np
from scipy import stats
from scipy.spatial.transform import Rotation

from avalon.common.utils import only
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.biome import generate_biome_config
from avalon.datagen.world_creation.configs.building import BuildingConfig
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import IndoorTaskConfig
from avalon.datagen.world_creation.configs.world import WorldConfig
from avalon.datagen.world_creation.constants import BOULDER_MAX_MASS
from avalon.datagen.world_creation.constants import BOULDER_MIN_MASS
from avalon.datagen.world_creation.constants import MAX_JUMP_HEIGHT_METERS
from avalon.datagen.world_creation.debug_plots import IS_DEBUG_VIS
from avalon.datagen.world_creation.entities.doors.hinge_door import HingeDoor
from avalon.datagen.world_creation.entities.doors.sliding_door import SlidingDoor
from avalon.datagen.world_creation.entities.entity import Entity
from avalon.datagen.world_creation.entities.food import CANONICAL_FOOD_CLASS
from avalon.datagen.world_creation.geometry import BuildingTile
from avalon.datagen.world_creation.geometry import local_to_global_coords
from avalon.datagen.world_creation.indoor.builders import DefaultEntranceBuilder
from avalon.datagen.world_creation.indoor.building import Building
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.indoor.components import Entrance
from avalon.datagen.world_creation.indoor.components import Room
from avalon.datagen.world_creation.indoor.components import Story
from avalon.datagen.world_creation.indoor.constants import DEFAULT_FLOOR_THICKNESS
from avalon.datagen.world_creation.indoor.constants import TILE_SIZE
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.doors import get_door_params_from_difficulty
from avalon.datagen.world_creation.indoor.doors import make_entrance_hinge_door
from avalon.datagen.world_creation.indoor.doors import make_entrance_sliding_door
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world import _get_spawn
from avalon.datagen.world_creation.worlds.world import build_building

EntranceSite = Tuple[int, int, Tuple[Azimuth, ...]]
IndoorTaskParams = Tuple[Building, Sequence[Entity], Point3DNP, Point3DNP, Tuple[EntranceSite, ...]]

CANONICAL_BUILDING_LOCATION = np.array([0.0, 2.0, 0.0])

PrincipalObstacleParams = TypeVar("PrincipalObstacleParams", bound=Tuple[Any, ...])


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class BuildingTaskGenerator(Generic[PrincipalObstacleParams]):
    config: IndoorTaskConfig = IndoorTaskConfig()

    def get_site_radius(self, rand: np.random.Generator, difficulty: float) -> float:
        raise NotImplementedError

    def get_building_config(
        self,
        rand: np.random.Generator,
        difficulty: float,
        radius: float,
        allowed_auxiliary_tasks: Tuple[BuildingTask, ...] = tuple(),
        aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig(),
    ) -> BuildingConfig:
        raise NotImplementedError

    def get_principal_obstacle_params(
        self, rand: np.random.Generator, difficulty: float, building: Building
    ) -> PrincipalObstacleParams:
        raise NotImplementedError

    def add_principal_obstacles(
        self,
        rand: np.random.Generator,
        building: Building,
        obstacle_params: PrincipalObstacleParams,
    ) -> IndoorTaskParams:
        raise NotImplementedError

    def get_entrance_obstacle_params(
        self,
        rand: np.random.Generator,
        difficulty: float,
        building: Building,
        entrance: Entrance,
        allowed_entrance_tasks: Tuple[BuildingTask, ...],
    ) -> Optional[Tuple]:
        if difficulty == 0:
            return None
        if BuildingTask.OPEN in allowed_entrance_tasks:
            return get_door_params_from_difficulty(rand, difficulty)
        return None

    def add_entrance(
        self, rand: np.random.Generator, building: Building, viable_entrance_sites: Tuple[EntranceSite, ...]
    ) -> None:
        site_idx = rand.choice(range(len(viable_entrance_sites)))
        story_id, room_id, permitted_azimuths = viable_entrance_sites[site_idx]
        entrance_builder = DefaultEntranceBuilder(top=False, bottom=True)
        entrance_builder.add_entrance(
            building.stories[story_id], rand, permitted_room_ids=[room_id], permitted_azimuths=permitted_azimuths
        )

    def add_entrance_obstacles(
        self, rand: np.random.Generator, building: Building, story: Story, entrance: Entrance, obstacle_params: Tuple
    ) -> Tuple[Building, List[Entity]]:
        door_type, door_mechanics_params, latching_mechanics, variant_locks = obstacle_params
        if door_type is HingeDoor:
            make_entrance_door = make_entrance_hinge_door
        elif door_type is SlidingDoor:
            make_entrance_door = make_entrance_sliding_door  # type: ignore
        else:
            raise NotImplementedError(door_type)

        door = make_entrance_door(
            story,
            entrance,
            latching_mechanics=latching_mechanics,
            **door_mechanics_params,
        )

        locks = []
        for make_lock, *params in variant_locks:
            params = only(params) if params else {}
            locks.append(make_lock(door, **params))
        with door.mutable_clone() as locked_door:
            locked_door.locks = locks
            door = locked_door

        return building, [door]

    @staticmethod
    def _position_from_tile(room: Room, tile: BuildingTile, at_height: float, tile_size: int = TILE_SIZE) -> Point3DNP:
        return np.array(
            [
                tile.x + tile_size / 2,
                room.floor_heightmap[tile.z, tile.x] + at_height,
                tile.z + tile_size / 2,
            ]
        )


def decide_spawn_room_and_tile(
    rand: np.random.Generator,
    story: Story,
    permitted_tiles_by_room_id: Optional[Dict[int, List[Tuple[int, int]]]] = None,
) -> Tuple[Room, Tuple[int, int]]:
    if entrances := story.entrances:
        entrance = rand.choice(entrances)  # type: ignore[arg-type]
        spawn_room, landing_position = entrance.get_connected_room_and_landing_position(story)
        spawn_tile = landing_position.x, landing_position.z
        if permitted_tiles_by_room_id is not None:
            if (
                spawn_room.id not in permitted_tiles_by_room_id
                or spawn_tile not in permitted_tiles_by_room_id[spawn_room.id]
            ):
                raise ImpossibleWorldError("Entrance was spawned in a location outside the spawn region")
    else:
        if permitted_tiles_by_room_id is not None:
            spawn_room_id = rand.choice(list(permitted_tiles_by_room_id.keys()))
            spawn_room = story.rooms[spawn_room_id]
            spawn_tile = tuple(rand.choice(permitted_tiles_by_room_id[spawn_room_id]))  # type: ignore
        else:
            spawn_room: Room = rand.choice(story.rooms)  # type: ignore
            spawn_tile = rand.integers(0, spawn_room.width - 1), rand.integers(0, spawn_room.length - 1)
    return spawn_room, spawn_tile


def _decide_boulder_mass(
    rand: np.random.Generator,
    difficulty: float,
    min_mass: float = BOULDER_MIN_MASS,
    max_mass: float = BOULDER_MAX_MASS,
    lighter_is_harder: bool = False,
) -> float:
    if lighter_is_harder:
        difficulty = 1 - difficulty
    target_mass = min_mass + (max_mass - min_mass) * difficulty
    mass_std = 2.5
    boulder_mass_distribution = stats.norm(target_mass, mass_std)
    return cast(float, min(max(boulder_mass_distribution.rvs(random_state=rand), min_mass), max_mass))


def add_food_island(
    room: Room, position: BuildingTile, size: int, height: float
) -> Tuple[Room, List[Tuple[int, int]]]:
    """
    Place an elevated square platform in the room with `position` as its top-left tile.
    Position should be passed in ROOM coordinate space.
    Returns all its tiles as (x,z) coordinates (in room space).
    """
    if size >= room.width or size >= room.length:
        raise ImpossibleWorldError(f"Room {room.id} is too small to place a food island of size {size}")

    new_heightmap = room.floor_heightmap.copy()
    new_heightmap[position.z : position.z + size, position.x : position.x + size] = DEFAULT_FLOOR_THICKNESS + height
    if height > 3 and size > 1:
        center_z = position.z + size // 2
        center_x = position.x + size // 2
        new_heightmap[center_z, center_x] += MAX_JUMP_HEIGHT_METERS * 0.95
    tiles = list(product(range(position.x, position.x + size), range(position.z, position.z + size)))
    new_outer_height = room.outer_height + (new_heightmap.max() - DEFAULT_FLOOR_THICKNESS)
    updated_room = attr.evolve(room, outer_height=new_outer_height, floor_heightmap=new_heightmap)
    return updated_room, tiles


def translate_entities_to_world_space(building: Building, entities: Sequence[Entity]):
    translated_entities = []
    for entity in entities:
        with entity.mutable_clone() as offset_entity:
            offset_entity.position = get_location_in_world_space(entity.position, building)
            if hasattr(offset_entity, "rotation") and building.yaw_degrees != 0:
                offset_entity.rotation = (
                    Rotation.from_euler("y", building.yaw_degrees, degrees=True).as_matrix().flatten()
                )
            translated_entities.append(offset_entity)
    return translated_entities


MIN_NAVIGATE_BUILDING_SIZE = 5.0


def rectangle_dimensions_within_radius(radius: float) -> Tuple[int, int]:
    size = math.floor(2 * radius / math.sqrt(2))
    return size, size


def get_room_centroid_in_building_space(
    building: Building, story: Story, room: Room, at_height: Optional[float] = None
) -> Point3DNP:
    tile_size = 1
    height = at_height if at_height else room.outer_height / 2
    story_y_offset = building.get_story_y_offset(story.num)
    center_tile_elevation = room.floor_heightmap[room.length // 2, room.width // 2]
    return np.array(
        [room.center.x + tile_size / 2, story_y_offset + center_tile_elevation + height, room.center.z + tile_size / 2]
    )


def get_room_centroid_in_world_space(
    building: Building, story: Story, room: Room, at_height: Optional[float] = None
) -> Point3DNP:
    centroid_in_building_space = get_room_centroid_in_building_space(building, story, room, at_height)
    centroid_in_world_space = local_to_global_coords(centroid_in_building_space, building.position)
    return centroid_in_world_space


def rebuild_with_aligned_entrance(
    building: Building, rand: np.random.Generator, initial_room_id: int, entrance_azimuth: Optional[Azimuth] = None
) -> Building:
    entrance_builder = DefaultEntranceBuilder(top=False, bottom=True)
    permitted_azimuths = [entrance_azimuth] if entrance_azimuth else None
    entrance = entrance_builder.add_entrance(
        building.stories[0], rand, permitted_room_ids=[initial_room_id], permitted_azimuths=permitted_azimuths
    )
    required_yaw_rotation = -entrance.azimuth.angle_from_positive_x
    return building.rebuild_rotated(required_yaw_rotation)


def make_indoor_task_world(
    building: Building,
    extra_entities: List[Entity],
    difficulty: float,
    spawn_location: Point3DNP,
    target_location: Point3DNP,
    rand: np.random.Generator,
    export_config: ExportConfig,
):
    if IS_DEBUG_VIS:
        building.plot()

    first_story = building.stories[0]
    furthest_distance = max(
        abs(building.position[0]) + first_story.width / 2, abs(building.position[2]) + first_story.length / 2
    )
    config = WorldConfig(
        seed=rand.integers(0, np.iinfo(np.int64).max),
        size_in_meters=2 * furthest_distance,
        point_density_in_points_per_square_meter=0.2,
        initial_point_count=1,
        initial_noise_scale=0,
        fractal_iteration_count=0,
        noise_scale_decay=0,
        mountain_noise_count=0,
        mountain_radius=0,
        mountain_offset=0,
        mountain_radius_decay=0,
        mountain_noise_scale=0,
        is_mountain_placement_normal_distribution=True,
        final_max_altitude_meters=0,
        is_indoor_only=True,
    )
    biome_config = generate_biome_config(rand, export_config, 10.0, 0.0)
    export_config = attr.evolve(export_config, is_biome_fast=True)
    world = World.build(config, export_config, biome_config, is_debug_graph_printing_enabled=False)
    building, world = world.add_building(building, np.ones_like(world.map.Z))
    world = world.add_items(
        [
            *extra_entities,
            _get_spawn(rand, difficulty, spawn_location, target_location),
            CANONICAL_FOOD_CLASS(position=target_location),
        ]
    )

    return world


def get_location_in_world_space(location_in_building: Point3DNP, building: Building):
    translated = local_to_global_coords(location_in_building, building.offset_point)
    offset_from_centroid = translated - building.position
    rotation = Rotation.from_euler("y", building.yaw_degrees, degrees=True)
    rotated = rotation.apply(offset_from_centroid)
    rotated_and_offset = rotated + building.position
    return rotated_and_offset


def create_building_for_skill_scenario(
    rand: np.random.Generator,
    difficulty: float,
    task_generator: BuildingTaskGenerator,
    position: Point3DNP,
    site_radius: Optional[float] = None,
    allowed_auxiliary_tasks: Tuple[BuildingTask, ...] = tuple(),
    allowed_entrance_tasks: Tuple[BuildingTask, ...] = tuple(),
    aesthetics_config: BuildingAestheticsConfig = BuildingAestheticsConfig(),
    is_indoor_only: bool = False,
):
    site_radius = task_generator.get_site_radius(rand, difficulty) if site_radius is None else site_radius

    default_building_id = 0
    building_config = task_generator.get_building_config(
        rand, difficulty, site_radius, allowed_auxiliary_tasks, aesthetics_config
    )
    building = build_building(building_config, default_building_id, rand)
    obstacle_params = task_generator.get_principal_obstacle_params(rand, difficulty, building)
    (
        building,
        entities,
        spawn_location_in_building,
        target_location_in_building,
        entrance_sites,
    ) = task_generator.add_principal_obstacles(rand, building, obstacle_params)

    if not is_indoor_only:
        task_generator.add_entrance(rand, building, entrance_sites)

        for story in building.stories:
            for entrance in story.entrances:
                add_entrance_obstacles = len(allowed_entrance_tasks) > 0 and rand.uniform() < difficulty
                if not add_entrance_obstacles:
                    continue

                entrance_obstacle_params = task_generator.get_entrance_obstacle_params(
                    rand, difficulty, building, entrance, allowed_entrance_tasks
                )
                if entrance_obstacle_params:
                    building, additional_entities = task_generator.add_entrance_obstacles(
                        rand, building, story, entrance, entrance_obstacle_params
                    )
                    entities = [*entities, *additional_entities]

    building = building.with_transform(new_position=position)
    spawn_location = get_location_in_world_space(spawn_location_in_building, building)
    target_location = get_location_in_world_space(target_location_in_building, building)
    translated_entities = translate_entities_to_world_space(building, entities)
    return building, translated_entities, spawn_location, target_location



from pathlib import Path
from typing import Tuple

import attr
import numpy as np
from networkx.algorithms.approximation import traveling_salesman_problem

from avalon.common.utils import only
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.building import BuildingConfig
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import IndoorTaskConfig
from avalon.datagen.world_creation.constants import AGENT_HEIGHT
from avalon.datagen.world_creation.entities.doors.hinge_door import HingeDoor
from avalon.datagen.world_creation.entities.doors.sliding_door import SlidingDoor
from avalon.datagen.world_creation.geometry import Axis
from avalon.datagen.world_creation.indoor.builders import HamiltonianHallwayBuilder
from avalon.datagen.world_creation.indoor.builders import HouseLikeRoomBuilder
from avalon.datagen.world_creation.indoor.builders import RectangleFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import WindowBuilder
from avalon.datagen.world_creation.indoor.builders import get_room_overlap
from avalon.datagen.world_creation.indoor.building import Building
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig
from avalon.datagen.world_creation.indoor.building import BuildingNavGraph
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.indoor.components import Room
from avalon.datagen.world_creation.indoor.components import Story
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.doors import DoorParams
from avalon.datagen.world_creation.indoor.doors import get_door_params_from_difficulty
from avalon.datagen.world_creation.indoor.doors import make_hinge_door
from avalon.datagen.world_creation.indoor.doors import make_sliding_door
from avalon.datagen.world_creation.indoor.task_generator import CANONICAL_BUILDING_LOCATION
from avalon.datagen.world_creation.indoor.task_generator import BuildingTaskGenerator
from avalon.datagen.world_creation.indoor.task_generator import IndoorTaskParams
from avalon.datagen.world_creation.indoor.task_generator import create_building_for_skill_scenario
from avalon.datagen.world_creation.indoor.task_generator import get_room_centroid_in_building_space
from avalon.datagen.world_creation.indoor.task_generator import make_indoor_task_world
from avalon.datagen.world_creation.indoor.task_generator import rectangle_dimensions_within_radius
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.export import export_world


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class OpenTaskConfig(IndoorTaskConfig):
    # Smallest and largest possible site radius for a building with this task to be placed on.
    min_site_radius: float = 7.0
    max_site_radius: float = 15.0
    # Total number of stories for the building. Note that this task is a single-story task, all other stories will
    # be purely aesthetic (e.g. to help with being viewed in the distance outdoors).
    story_count: int = 2
    # Max number of rooms that will be created at difficulty=1. We'll have as many doors as rooms minus 1.
    # Note that going higher than 5 will increase number of impossible worlds, as it becomes harder to find a
    # Hamiltonian path in the building.
    max_room_count: int = 5


def generate_open_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: OpenTaskConfig = OpenTaskConfig(),
) -> None:
    building, entities, spawn_location, target_location = create_building_for_skill_scenario(
        rand,
        difficulty,
        OpenTaskGenerator(task_config),
        position=CANONICAL_BUILDING_LOCATION,
        is_indoor_only=True,
    )
    world = make_indoor_task_world(
        building, entities, difficulty, spawn_location, target_location, rand, export_config
    )
    export_world(output_path, rand, world)


StoryNum = int
FromRoom = int
ToRoom = int
PlacedDoorParams = Tuple[StoryNum, FromRoom, ToRoom, DoorParams]

FromStoryNum = int
FromRoomId = int
ToStoryNum = int
ToRoomId = int
OpenTaskObstacleParams = Tuple[FromStoryNum, FromRoomId, ToStoryNum, ToRoomId, Tuple[PlacedDoorParams, ...]]


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class OpenTaskGenerator(BuildingTaskGenerator):
    config: OpenTaskConfig = OpenTaskConfig()

    def get_site_radius(self, rand: np.random.Generator, difficulty: float) -> float:
        return scale_with_difficulty(difficulty, self.config.min_site_radius, self.config.max_site_radius)

    def get_building_config(
        self,
        rand: np.random.Generator,
        difficulty: float,
        radius: float,
        allowed_auxiliary_tasks: Tuple[BuildingTask, ...] = tuple(),
        aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig(),
    ) -> BuildingConfig:
        width, length = rectangle_dimensions_within_radius(radius)
        if width < 5 and length < 5:
            raise ImpossibleWorldError("Building too small to fit open task")

        min_room_count = 2
        max_room_count = min_room_count + round(
            normal_distrib_range(0, self.config.max_room_count - min_room_count, 0.1, rand, difficulty)
        )

        return BuildingConfig(
            width=width,
            length=length,
            story_count=self.config.story_count,
            footprint_builder=RectangleFootprintBuilder(),
            # Min room size must be >3 to ensure we have space for locks on both sides
            room_builder=HouseLikeRoomBuilder(min_room_size=3, max_rooms=max_room_count),
            hallway_builder=HamiltonianHallwayBuilder(),
            # All the action takes place on the first story; any extra stories are purely aesthetic and not connected
            story_linker=None,
            window_builder=WindowBuilder(),
            aesthetics=aesthetics,
        )

    def get_principal_obstacle_params(
        self, rand: np.random.Generator, difficulty: float, building: Building
    ) -> OpenTaskObstacleParams:
        story_num = 0
        if len(building.stories[story_num].rooms) < 2:
            raise ImpossibleWorldError("Can't make open task in a single room")
        decorative_story_nums = tuple(s.num for s in building.stories[story_num + 1 :])
        nav_graph = BuildingNavGraph(building, excluded_stories=decorative_story_nums)
        path = traveling_salesman_problem(nav_graph, cycle=False)
        all_door_params = []
        stories_and_rooms_by_node_id = nav_graph.get_stories_and_rooms_by_node_id(building.stories)
        for from_node, to_node in zip(path[:-1], path[1:]):
            from_story, from_room = stories_and_rooms_by_node_id[from_node]
            to_story, to_room = stories_and_rooms_by_node_id[to_node]
            self._check_connection_is_valid(from_story, from_room, to_story, to_room)
            all_door_params.append(
                (story_num, from_room.id, to_room.id, get_door_params_from_difficulty(rand, difficulty))
            )
        initial_story, initial_room = stories_and_rooms_by_node_id[path[0]]
        target_story, target_room = stories_and_rooms_by_node_id[path[-1]]
        return initial_story.num, initial_room.id, target_story.num, target_room.id, tuple(all_door_params)

    @staticmethod
    def _check_connection_is_valid(from_story: Story, from_room: Room, to_story: Story, to_room: Room) -> None:
        # Sometimes hallways can only be fit in the corner of rooms - this does not work for the Open task since locks
        # must be placed on the sides and there might not be enough room.
        if from_story == to_story:
            x_overlap, z_overlap = get_room_overlap(from_room, to_room)
            wall_overlap = x_overlap or z_overlap
            is_hallway_going_to_be_forced_into_corner = wall_overlap is not None and wall_overlap.size < 3
            if is_hallway_going_to_be_forced_into_corner:
                raise ImpossibleWorldError(
                    f"Wall overlap between {from_room.id} and {to_room.id} is too small to fit a door with locks properly."
                )

    def add_principal_obstacles(
        self, rand: np.random.Generator, building: Building, obstacle_params: OpenTaskObstacleParams
    ) -> IndoorTaskParams:
        initial_story_num, initial_room_id, target_story_num, target_room_id, placed_door_param_set = obstacle_params
        from_story = building.stories[initial_story_num]
        from_room = from_story.rooms[initial_room_id]
        spawn_location = get_room_centroid_in_building_space(
            building, from_story, from_room, at_height=AGENT_HEIGHT / 2
        )

        to_story = building.stories[target_story_num]
        to_room = to_story.rooms[target_room_id]
        target_location = get_room_centroid_in_building_space(building, to_story, to_room, at_height=AGENT_HEIGHT / 2)

        doors = []
        for story_num, from_room_id, to_room_id, door_params in placed_door_param_set:
            story = building.stories[story_num]
            from_room = story.rooms[from_room_id]
            to_room = story.rooms[to_room_id]
            hallway = only(
                [
                    hallway
                    for hallway in story.hallways
                    if {from_room.id, to_room.id} == {hallway.from_room_id, hallway.to_room_id}
                ]
            )

            if hallway.from_room_id == from_room.id:
                door_tile = hallway.points[0]
                door_azimuth = hallway.from_room_azimuth
            else:
                door_tile = hallway.points[-1]
                door_azimuth = hallway.to_room_azimuth

            if door_azimuth in {Azimuth.NORTH, Azimuth.SOUTH}:
                door_face_axis = Axis.X
            else:
                door_face_axis = Axis.Z

            door_type, door_mechanics_params, latching_mechanics, variant_locks = door_params
            if door_type is HingeDoor:
                make_door = make_hinge_door
            elif door_type is SlidingDoor:
                make_door = make_sliding_door  # type: ignore
            else:
                raise NotImplementedError(door_type)

            door = make_door(
                story,
                door_tile,
                door_azimuth,
                door_face_axis,
                latching_mechanics=latching_mechanics,
                **door_mechanics_params,
            )

            locks = []
            for make_lock, *params in variant_locks:
                final_params = only(params) if params else {}
                locks.append(make_lock(door, **final_params))  # type: ignore
            with door.mutable_clone() as locked_door:
                locked_door.locks = locks
                door = locked_door
            doors.append(door)

        entrance_sites = tuple([(initial_story_num, initial_room_id, tuple(Azimuth))])
        return building, doors, spawn_location, target_location, entrance_sites



import attr
import numpy as np
from godot_parser import Node as GDNode

from avalon.common.utils import only
from avalon.datagen.world_creation.constants import IDENTITY_BASIS
from avalon.datagen.world_creation.entities.doors.locks.door_lock import DoorLock
from avalon.datagen.world_creation.indoor.blocks import new_make_block_nodes
from avalon.datagen.world_creation.indoor.constants import DEFAULT_STORY_HEIGHT
from avalon.datagen.world_creation.indoor.utils import get_scaled_mesh_node
from avalon.datagen.world_creation.types import GodotScene


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class DoorOpenButton(DoorLock):
    size: np.ndarray = np.array([1, DEFAULT_STORY_HEIGHT, 0.1])
    rotation: np.ndarray = IDENTITY_BASIS

    def get_node(self, scene: GodotScene) -> GDNode:
        button_mesh_node = get_scaled_mesh_node(scene, "res://entities/doors/open_button.tscn", self.size)
        button_node = only(
            new_make_block_nodes(
                scene,
                position=self.position,
                size=self.size,
                rotation=self.rotation,
                make_parent=True,
                parent_type="RigidBody",
                parent_name=f"open_button_{self.entity_id}",
                parent_script=scene.add_ext_resource("res://entities/doors/open_button.gd", "Script").reference,
                parent_extra_props={
                    "entity_id": self.entity_id,
                },
                make_mesh=False,
                make_collision_shape=True,
            )
        )
        button_node.add_child(button_mesh_node)
        return button_node



import math
from enum import Enum
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import cast

import attr
import networkx as nx
import numpy as np
from loguru import logger
from nptyping import assert_isinstance
from scipy import stats
from skimage import morphology

from avalon.common.errors import SwitchError
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.building import generate_aesthetics_config
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import IndoorTaskConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import ITEM_FLATTEN_RADIUS
from avalon.datagen.world_creation.constants import UP_VECTOR
from avalon.datagen.world_creation.constants import WATER_LINE
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.debug_plots import plot_terrain
from avalon.datagen.world_creation.debug_plots import plot_value_grid
from avalon.datagen.world_creation.entities.animals import ALL_PREDATOR_CLASSES
from avalon.datagen.world_creation.entities.animals import Predator
from avalon.datagen.world_creation.entities.constants import CANONICAL_FOOD_HEIGHT_ON_TREE
from avalon.datagen.world_creation.entities.constants import FOOD_TREE_VISIBLE_HEIGHT
from avalon.datagen.world_creation.entities.entity import Entity
from avalon.datagen.world_creation.entities.food import CANONICAL_FOOD_CLASS
from avalon.datagen.world_creation.entities.tools.placeholder import Placeholder
from avalon.datagen.world_creation.entities.tools.weapons import LargeRock
from avalon.datagen.world_creation.entities.tools.weapons import Rock
from avalon.datagen.world_creation.entities.tools.weapons import Stick
from avalon.datagen.world_creation.entities.tools.weapons import Weapon
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.indoor.building import get_building_footprint_outline
from avalon.datagen.world_creation.indoor.constants import DEFAULT_FLOOR_THICKNESS
from avalon.datagen.world_creation.indoor.task_generator import MIN_NAVIGATE_BUILDING_SIZE
from avalon.datagen.world_creation.indoor.task_generator import BuildingTaskGenerator
from avalon.datagen.world_creation.indoor.task_generator import create_building_for_skill_scenario
from avalon.datagen.world_creation.tasks.bridge import BridgeTaskConfig
from avalon.datagen.world_creation.tasks.bridge import create_bridge_obstacle
from avalon.datagen.world_creation.tasks.climb import ClimbIndoorTaskConfig
from avalon.datagen.world_creation.tasks.climb import ClimbOutdoorTaskConfig
from avalon.datagen.world_creation.tasks.climb import ClimbTaskGenerator
from avalon.datagen.world_creation.tasks.climb import create_climb_obstacle
from avalon.datagen.world_creation.tasks.descend import DescendTaskConfig
from avalon.datagen.world_creation.tasks.descend import create_descend_obstacle
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.tasks.explore import ExploreIndoorTaskConfig
from avalon.datagen.world_creation.tasks.explore import ExploreTaskGenerator
from avalon.datagen.world_creation.tasks.fight import FightTaskConfig
from avalon.datagen.world_creation.tasks.fight import create_fight_obstacle
from avalon.datagen.world_creation.tasks.jump import JumpIndoorTaskConfig
from avalon.datagen.world_creation.tasks.jump import JumpOutdoorTaskConfig
from avalon.datagen.world_creation.tasks.jump import JumpTaskGenerator
from avalon.datagen.world_creation.tasks.jump import create_jump_obstacle
from avalon.datagen.world_creation.tasks.move import MoveTaskConfig
from avalon.datagen.world_creation.tasks.move import create_move_obstacle
from avalon.datagen.world_creation.tasks.open import OpenTaskConfig
from avalon.datagen.world_creation.tasks.open import OpenTaskGenerator
from avalon.datagen.world_creation.tasks.push import PushIndoorTaskConfig
from avalon.datagen.world_creation.tasks.push import PushOutdoorTaskConfig
from avalon.datagen.world_creation.tasks.push import PushTaskGenerator
from avalon.datagen.world_creation.tasks.push import create_push_obstacle
from avalon.datagen.world_creation.tasks.stack import StackIndoorTaskConfig
from avalon.datagen.world_creation.tasks.stack import StackOutdoorTaskConfig
from avalon.datagen.world_creation.tasks.stack import StackTaskGenerator
from avalon.datagen.world_creation.tasks.stack import create_stack_obstacle
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import Point2DNP
from avalon.datagen.world_creation.types import Point3DListNP
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.types import WorldType
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.creation import create_world_for_skill_scenario
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_categorical_difficulty
from avalon.datagen.world_creation.worlds.height_map import HeightMap
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations

# just need to be tuned for performance. Basically, more retries makes it more likely to generate a world with enough
# obstacles in worlds that are tricky but barely possible, while too many retries means you waste time continuing to
# try to place obstacles in worlds where it's basically impossible to do so and you just need to give up and start over
_OUTDOOR_OBSTACLE_MAX_RETRIES = 4
_INDOOR_OBSTACLE_MAX_RETRIES = 5


OUTDOOR_OBSTACLE_CONFIG_DICT: Dict[AvalonTask, Tuple[float, TaskConfig]] = {
    AvalonTask.MOVE: (1.0, MoveTaskConfig()),
    AvalonTask.JUMP: (1.0, JumpOutdoorTaskConfig()),
    AvalonTask.CLIMB: (1.0, ClimbOutdoorTaskConfig()),
    AvalonTask.PUSH: (1.0, PushOutdoorTaskConfig()),
    AvalonTask.FIGHT: (1.0, FightTaskConfig()),
    AvalonTask.STACK: (1.0, StackOutdoorTaskConfig()),
    AvalonTask.BRIDGE: (1.0, BridgeTaskConfig()),
    AvalonTask.DESCEND: (1.0, DescendTaskConfig()),
}

INDOOR_OBSTACLE_CONFIG_DICT: Dict[BuildingTask, Tuple[float, IndoorTaskConfig]] = {
    BuildingTask.EXPLORE: (1.0, ExploreIndoorTaskConfig()),
    BuildingTask.OPEN: (1.0, OpenTaskConfig()),
    BuildingTask.PUSH: (1.0, PushIndoorTaskConfig()),
    BuildingTask.STACK: (1.0, StackIndoorTaskConfig()),
    BuildingTask.CLIMB: (1.0, ClimbIndoorTaskConfig()),
    BuildingTask.JUMP: (1.0, JumpIndoorTaskConfig()),
}


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class CompositeTaskConfig(TaskConfig):
    # set this to the particular task being done
    task: AvalonTask
    # the largest number of obstacles that will ever be put in a single world. much more than 4 is pretty annoying to
    # play, and starts to require relatively large levels (if they're all height obstacles).
    # Higher difficulties will have more obstacles (up to this number)
    max_obstacles: int = 4
    # how frequently to include more or fewer obstacles (true obstacle count is sampled from:
    # NormalDistribution(max_obstacles, max_obstacles_std_dev)
    max_obstacles_std_dev: float = 1.0
    # how much world size is required for each obstacle. Rough number based on the max required by any individual
    # obstacle. Is required to determine how large the world should be
    goal_distance_per_obstacle: float = 21.0
    # how much to vary the size of the world. Worlds must be large enough so that even a two sigma small world
    # will still be big enough to fit all obstacles
    world_size_std_dev: float = 20.0
    # how high to boost the spawn to ensure that the goal is visible. If boosting up the spawn this much doesn't
    # result in the goal being visible, but it does need to be visible, an ImpossibleWorldError will be raised
    max_visibility_height_boost: float = 8.0
    # the probability that the food ends up in a building (which can have its own obstacles)
    is_goal_in_building_probability: float = 0.5
    # Compositional height obstacles present a problem: if the world is a normal island, you can just walk around
    # many of the obstacles. There are two different ways of dealing with this issue:
    # 1. raise the islands out of the water such that all land is surrounded by unclimbable cliffs
    # 2. only use obstalces that do not have this issue
    # This parameter controls the probability of the first mode. It is made more common because that way we get to use
    # most of the obstacles. The second mode has a very particular look because it can really only use a small number
    # of the obstacles (climb, descend, etc--things that change height)
    is_raised_island_mode_probability: float = 0.8
    # how many predators to include. At difficulty 1.0, will be a value randomly chosen between the two extremes
    predator_count_easy: int = 0
    predator_count_hard: int = 16
    # don't start predators within this radius of where you start
    min_predator_distance_from_spawn: float = 10.0
    #
    predator_path_dist_easy: float = 40.0
    predator_path_dist_hard: float = 15.0
    #
    weapon_ratio_easy: float = 3.0
    weapon_ratio_hard: float = 0.0
    weapon_path_dist_easy: float = 5.0
    weapon_path_dist_hard: float = 20.0
    large_rock_probability: float = 0.4
    rock_probability_easy: float = 0.5
    rock_probability_hard: float = 0.95
    # you can configure any of the inner tasks, as well as the probability of each of them occuring (will be normalized)
    outdoor_obstacle_configs: Dict[AvalonTask, Tuple[float, TaskConfig]] = OUTDOOR_OBSTACLE_CONFIG_DICT
    # you can also configure any of the indoor tasks that are used as well
    indoor_obstacle_configs: Dict[BuildingTask, Tuple[float, IndoorTaskConfig]] = INDOOR_OBSTACLE_CONFIG_DICT


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ForcedComposition:
    is_enabled: bool = False


# These are all the obstacles that create rings around a point
_ALLOWED_OBSTACLES = [
    AvalonTask.MOVE,
    AvalonTask.JUMP,
    AvalonTask.CLIMB,
    AvalonTask.PUSH,
    AvalonTask.FIGHT,
    AvalonTask.STACK,
    AvalonTask.BRIDGE,
    AvalonTask.DESCEND,
]


class FinalObstacleType(Enum):
    PREDATOR = "PREDATOR"
    CARRY = "CARRY"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class CompositionalSolutionPath:
    mask: MapBoolNP
    path_locations: np.ndarray
    points: Point3DListNP
    point_indices: np.ndarray
    start_2d_radii: np.ndarray
    end_2d_radii: np.ndarray

    def get_point_at_radius_from_goal(self, radius: float) -> Point3DNP:
        far_enough_away = self.end_2d_radii > radius
        if not np.any(far_enough_away):
            raise ImpossibleWorldError("The safety radius for the end point included the start point!")
        idx = int(np.argwhere(far_enough_away)[-1])
        return cast(Point3DNP, self.points[idx])

    def get_point_at_radius_from_spawn(self, radius: float) -> Point3DNP:
        far_enough_away = self.start_2d_radii > radius
        if not np.any(far_enough_away):
            raise ImpossibleWorldError("The safety radius for the start point included the end point!")
        idx = int(np.argwhere(far_enough_away)[0])
        return cast(Point3DNP, self.points[idx])


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class CompositionalSolutionData:
    locations: WorldLocations
    full_path: CompositionalSolutionPath
    current_spawn_side_radius: float = 0.0
    current_goal_side_radius: float = 0.0

    @staticmethod
    def build(
        locations: WorldLocations, full_path: CompositionalSolutionPath, goal_safety_radius: float
    ) -> "CompositionalSolutionData":
        return CompositionalSolutionData(
            locations=locations,
            full_path=full_path,
            current_goal_side_radius=goal_safety_radius,
        )

    @property
    def spawn_endpoint(self) -> Point3DNP:
        return self.full_path.get_point_at_radius_from_spawn(self.current_spawn_side_radius)

    @property
    def goal_endpoint(self) -> Point3DNP:
        return self.full_path.get_point_at_radius_from_goal(self.current_goal_side_radius)

    def get_distance_to_use(self, remaining_height_obstacles: int) -> float:
        how_much_distance_to_use = 1 / remaining_height_obstacles
        distance_left = (
            self.locations.get_2d_spawn_goal_distance()
            - (self.current_goal_side_radius + self.current_spawn_side_radius) * 1.1
        )
        return how_much_distance_to_use * distance_left

    def update_spawn(self, radius: float) -> "CompositionalSolutionData":
        return attr.evolve(self, current_spawn_side_radius=radius)

    def update_goal(self, radius: float) -> "CompositionalSolutionData":
        return attr.evolve(self, current_goal_side_radius=radius)

    def plot(self, map: HeightMap) -> None:
        map.plot()
        heights = map.Z.copy()
        max_height = np.max(map.Z)
        heights[map.Z < WATER_LINE] = max_height
        interesting_points = [self.locations.spawn, self.locations.goal, self.spawn_endpoint, self.goal_endpoint]
        markers = [map.point_to_index(np.array([x[0], x[2]])) for x in interesting_points]
        for location, p in zip(list(self.full_path.path_locations), list(self.full_path.point_indices)):
            # if location < self.goal_path_location and location > self.spawn_path_location:
            #     heights[tuple(p)] = max_height
            heights[tuple(p)] = max_height
        plot_value_grid(heights, "Current solution", markers=markers)


def is_building_too_close_to_map_edge(building_center: Point2DNP, radius: float, map: HeightMap) -> bool:
    for delta in [
        (-radius, 0.0),
        (radius, 0.0),
        (0.0, -radius),
        (0.0, radius),
    ]:
        if not map.region.contains_point_2d(building_center + np.array(delta)):
            return True
    return False


def create_compositional_task(
    rand: np.random.Generator,
    difficulty: float,
    task_config: CompositeTaskConfig,
    export_config: ExportConfig,
    desired_goal_distance: Optional[stats.norm] = None,
    is_food_allowed: bool = True,
    is_debug_graph_printing_enabled: bool = False,
    min_size_in_meters: Optional[float] = None,
    max_size_in_meters: Optional[float] = None,
    _FORCED: Optional[ForcedComposition] = None,
) -> Tuple[World, WorldLocations]:
    """
    how this works, at a high level:

    first we use difficulty figure out how many obstacles we want, and how far to travel

    then we figure out how many of each type of obstacle we want
    obstacles come in three types:
        buildings
        outdoor rings
        final modifiers

    if the destination is a building, that is handled first.

    next we hand the ring obstacles, which are the most complicated to add

    first we try to add single small rings around the spawn and goal points (if desired)

    then we add the rest of the ring obstacles along the path inbetween.
    the path needs to be calculated so that we can move from the start to the end, staying on the land
    whenever a ring obstacle is added, we effectively are consuming part of that path

    after the ring obstacles are added, we add the buildings

    at the very end, we try to add enough final modifiers to bring us up to our desired total
    """

    # TODO: put back
    is_debug_graph_printing_enabled = False

    desired_obstacle_count, difficulty = _get_desired_obstacle_count(rand, difficulty, task_config, _FORCED)
    # figure out how many of each type of obstacle we want
    (building_count, desired_height_obstacle_count) = _get_obstacle_counts_per_type(
        rand, desired_obstacle_count, task_config, is_food_allowed
    )
    if task_config.task == AvalonTask.SURVIVE:
        # this is because it is handled outside of this function
        random_predator_count = 0
    else:
        random_predator_count = round(
            difficulty_variation(
                task_config.predator_count_easy,
                task_config.predator_count_hard,
                rand,
                difficulty,
            )
        )

    if desired_goal_distance is None:
        desired_goal_distance = (
            desired_obstacle_count * task_config.goal_distance_per_obstacle + 2 * task_config.world_size_std_dev
        )

    if max_size_in_meters is None:
        # 2x because we want space for the obstacles, and additional 1x to account fo the 50%
        # variation in world size from the way the fractal stuff works out
        max_size_in_meters = 3.0 * desired_goal_distance

        if building_count > 0:
            max_size_in_meters += 60.0

    if task_config.task == AvalonTask.NAVIGATE:
        is_visibility_required = True
    else:
        is_visibility_required = False

    if is_debug_graph_printing_enabled:
        logger.debug(
            f"Obstacles: {building_count} buildings, {desired_height_obstacle_count} height, {random_predator_count} predators"
        )
        logger.debug(f"Visibility: {is_visibility_required}")

    is_destination_building = False
    if task_config.task in (AvalonTask.GATHER, AvalonTask.SURVIVE):
        is_destination_building = building_count > 0
    elif building_count > 0:
        is_destination_building = True
        assert (
            building_count <= 1
        ), "Can only create a single food, and buildings must have food, thus cannot create extra buildings"

    min_visible_height_for_world_creation = FOOD_TREE_VISIBLE_HEIGHT
    if is_destination_building:
        # this is roughly the height of the smallest building
        min_visible_height_for_world_creation = 4.0

    goal_distance = stats.norm(desired_goal_distance, task_config.world_size_std_dev)
    world, starting_locations = create_world_for_skill_scenario(
        rand,
        difficulty,
        CANONICAL_FOOD_HEIGHT_ON_TREE,
        goal_distance,
        export_config,
        is_visibility_required=is_visibility_required,
        min_size_in_meters=min_size_in_meters,
        max_size_in_meters=max_size_in_meters,
        visibility_height=min_visible_height_for_world_creation,
        world_type=WorldType.CONTINENT,
    )

    if is_debug_graph_printing_enabled:
        logger.debug(f"WORLD SIZE: {world.config.size_in_meters ** 2}")
        logger.debug(
            f"Goal and spawn are {starting_locations.get_2d_spawn_goal_distance()} away, we need at least {desired_obstacle_count * task_config.goal_distance_per_obstacle}"
        )

    # we're going to keep adding obstacles until we hit our count
    obstacle_count = 0

    # generate a single AestheticsConfig that is shared among the buildings in the world
    aesthetics = generate_aesthetics_config()

    goal_safety_radius = 0.0
    is_food_required = True
    # you would expect this, but actually, we just care about being able to see to the top of the tree
    # goal_visible_height_delta = CANONICAL_FOOD_HEIGHT_ON_TREE
    goal_visible_height_delta = FOOD_TREE_VISIBLE_HEIGHT
    initial_building = None
    initial_building_items = None
    initial_building_starting_terrain_height = None
    if is_destination_building:
        building_task = _select_building_obstacle_type(rand, task_config)
        building_radius = get_radius_for_building_task(rand, building_task, difficulty, task_config)
        building_location_2d = to_2d_point(starting_locations.goal)
        if is_building_too_close_to_map_edge(building_location_2d, building_radius, world.map):
            building_obstacle = get_random_building_obstacle(rand, world, task_config, starting_locations, difficulty)
            if building_obstacle:
                building_task, building_radius, mask, new_difficulty = building_obstacle
                building_location_2d = world.map.index_to_point_2d(tuple(rand.choice(np.argwhere(mask))))  # type: ignore[arg-type]
            else:
                raise ImpossibleWorldError("Can not place initial building")
        new_goal_location, initial_building, initial_building_items, world = create_building(
            rand,
            world,
            task_config,
            building_location_2d,
            building_radius,
            building_task,
            difficulty,
            aesthetics,
            is_shore_created=True,
        )
        initial_building_starting_terrain_height = world.map.get_rough_height_at_point(building_location_2d)
        is_food_required = False
        # have to reset the island mask as well because the land might have been extended
        starting_locations = attr.evolve(
            starting_locations,
            goal=new_goal_location,
            island=world.map.get_island(to_2d_point(starting_locations.spawn))[-1],
        )
        obstacle_count += 1
        building_count -= 1
        if obstacle_count >= desired_obstacle_count:
            # just add the items now because we're done
            world = world.add_items(initial_building_items)
            world = world.add_item(CANONICAL_FOOD_CLASS(position=new_goal_location))
            full_path = _find_path(rand, world, starting_locations)
            solution = CompositionalSolutionData.build(starting_locations, full_path, goal_safety_radius)
            world = add_extra_predators(
                rand, world, task_config, solution, difficulty, random_predator_count, full_path
            )

            return _return_data(
                rand, world, difficulty, starting_locations, is_food_required, is_visibility_required, task_config
            )

        goal_safety_radius = building_radius * 1.5
        # this is how much above the goal is visible
        goal_visible_height_delta = initial_building.height.max_lt - starting_locations.goal

    # figure out how we're going to move from the start to the end
    full_path = _find_path(rand, world, starting_locations)
    solution = CompositionalSolutionData.build(starting_locations, full_path, goal_safety_radius)

    if is_debug_graph_printing_enabled:
        logger.debug(f"Initial safety radius: {goal_safety_radius}")
        logger.debug(f"Reaming distance: {solution.locations.get_2d_spawn_goal_distance()}")

    if is_debug_graph_printing_enabled:
        solution.plot(world.map)
        # plot_value_grid(
        #     world.map.Z,
        #     markers=[
        #         # world.map.point_to_index(to_2d_point(solution.full_path.get_point_at_radius_from_goal(x)))
        #         world.map.point_to_index(to_2d_point(solution.full_path.get_point_at_radius_from_spawn(x)))
        #         for x in range(50)
        #     ],
        # )

    if is_debug_graph_printing_enabled:
        if is_visibility_required:
            logger.debug("Visibility required!!!")
        logger.debug("Selected obstacle types:")
        logger.debug(f"{building_count + int(is_destination_building)} buildings")
        logger.debug(f"{desired_height_obstacle_count} height obstacles")
        logger.debug(f"{random_predator_count} predators")

    # raise the island up
    if desired_height_obstacle_count > 0:
        new_locations, world = world.begin_height_obstacles(solution.locations)
        solution = attr.evolve(solution, locations=new_locations)

        # see docs for is_raised_island_mode_probability
        is_raised_island_mode = rand.uniform() < task_config.is_raised_island_mode_probability

        if is_debug_graph_printing_enabled:
            logger.debug(f"{is_raised_island_mode=}")
        if is_raised_island_mode:
            # add as many ring obstacles as we can (up to the max we want)

            # keep track of which later obstacles are allowed for navigate, since we need to see the goal when swe start
            is_boosted = False
            if is_visibility_required:
                allowed_obstacles = tuple(
                    [
                        x
                        for x in _ALLOWED_OBSTACLES
                        if x
                        not in (
                            AvalonTask.DESCEND,
                            AvalonTask.CLIMB,
                            AvalonTask.FIGHT,
                        )
                    ]
                )
                is_boosted = rand.uniform() > 0.5
            else:
                allowed_obstacles = tuple([x for x in _ALLOWED_OBSTACLES])

            distance_to_use = solution.get_distance_to_use(desired_height_obstacle_count)
            spawn_region = None

            if is_boosted:
                obstacle_type = AvalonTask.CLIMB
                result, new_difficulty, solution, new_locations, world = _add_height_obstacle(
                    rand,
                    difficulty,
                    world,
                    obstacle_type,
                    task_config,
                    solution,
                    distance_to_use,
                    is_around_spawn=False,
                    is_height_inverted=True,
                    is_debug_graph_printing_enabled=is_debug_graph_printing_enabled,
                )
                if result:
                    desired_height_obstacle_count -= 1
                    allowed_obstacles += tuple([AvalonTask.CLIMB, AvalonTask.FIGHT])
                    # move the food to the ledge
                    if is_food_required:
                        updated_locations = attr.evolve(solution.locations, goal=new_locations.goal)
                        solution = attr.evolve(solution, locations=updated_locations)

                    # figure out the spawn region (in case this is the only obstacle)
                    spawn_region = world.obstacle_zones[0][0]

            is_first_successful_obstacle = True
            for i in range(desired_height_obstacle_count):
                obstacle_type = _select_ring_obstacle_type(rand, allowed_obstacles, task_config)
                is_around_spawn = True
                result, new_difficulty, solution, new_locations, world = _add_height_obstacle(
                    rand,
                    difficulty,
                    world,
                    obstacle_type,
                    task_config,
                    solution,
                    distance_to_use,
                    is_around_spawn=is_around_spawn,
                    is_height_inverted=False,
                    is_debug_graph_printing_enabled=is_debug_graph_printing_enabled,
                )
                # logger.debug(f"goal height {solution.locations.goal[1]}")
                if result:
                    obstacle_count += 1
                    if is_first_successful_obstacle:
                        is_first_successful_obstacle = False
                        # reset the spawn region
                        spawn_region = world.obstacle_zones[-1][0]

            if spawn_region is None:
                raise ImpossibleWorldError("Failed to add any height obstacles in raised island mode")
            world, new_locations = world.end_height_obstacles(
                solution.locations, is_accessible_from_water=False, spawn_region=spawn_region
            )
            solution = attr.evolve(solution, locations=new_locations)

        else:
            # add ring based obstacle around spawn
            fall_count, climb_count = _get_fall_climb_split(rand, desired_height_obstacle_count)

            is_first_successful_obstacle = True
            spawn_region = None

            distance_to_use = solution.get_distance_to_use(desired_height_obstacle_count)
            for i in range(fall_count):
                obstacle_type = AvalonTask.DESCEND
                result, new_difficulty, solution, new_locations, world = _add_height_obstacle(
                    rand,
                    difficulty,
                    world,
                    obstacle_type,
                    task_config,
                    solution,
                    distance_to_use,
                    is_around_spawn=True,
                    is_height_inverted=False,
                    is_debug_graph_printing_enabled=is_debug_graph_printing_enabled,
                )
                if result:
                    obstacle_count += 1
                    if is_first_successful_obstacle:
                        is_first_successful_obstacle = False
                        # reset the spawn region
                        spawn_region = world.obstacle_zones[-1][0]

            # add ring based obstacle around goal
            for i in range(climb_count):
                obstacle_type = AvalonTask.CLIMB
                result, new_difficulty, solution, new_locations, world = _add_height_obstacle(
                    rand,
                    difficulty,
                    world,
                    obstacle_type,
                    task_config,
                    solution,
                    distance_to_use,
                    is_around_spawn=False,
                    is_height_inverted=True,
                    is_debug_graph_printing_enabled=is_debug_graph_printing_enabled,
                )
                if result:
                    obstacle_count += 1
                    if is_first_successful_obstacle:
                        is_first_successful_obstacle = False
                        # reset the spawn region
                        spawn_region = world.obstacle_zones[-1][0]

            if spawn_region is None:
                raise ImpossibleWorldError("Failed to add any height obstacles in flat island mode")
            world, new_locations = world.end_height_obstacles(
                solution.locations, is_accessible_from_water=True, spawn_region=spawn_region
            )
            solution = attr.evolve(solution, locations=new_locations)

        # if necessary, boost JUST the spawn safety radius (until the end is visible)
        if is_visibility_required:
            height_increments = 0
            max_increments = 20
            height_delta_per_step = task_config.max_visibility_height_boost / max_increments
            visibility_height_offset = 0.0
            visibility_calculator = world.map.generate_visibility_calculator()
            # technically should add half agent offset to the first part, but want to be sure we can see it
            while not visibility_calculator.is_visible_from(
                solution.locations.spawn + UP_VECTOR * visibility_height_offset,
                solution.locations.goal + UP_VECTOR * goal_visible_height_delta,
            ):
                height_increments += 1
                if height_increments > max_increments:
                    # interesting_points = [solution.locations.spawn, solution.locations.goal]
                    # interesting_points = [solution.locations.goal]
                    # markers = [world.map.point_to_index(np.array([x[0], x[2]])) for x in interesting_points]
                    #
                    # logger.debug("busted")
                    # logger.debug((
                    #     solution.locations.spawn + up_vector * visibility_height_offset,
                    #     solution.locations.goal + up_vector * min_food_offset,
                    # ))
                    # logger.debug("works?")
                    # vis = world.plot_visibility(
                    #     solution.locations.spawn + up_vector * visibility_height_offset,
                    #     up_vector * min_food_offset,
                    #     markers,
                    # )
                    # world.plot_visibility(
                    #     locations.spawn + up_vector * visibility_height_offset + HALF_AGENT_HEIGHT_VECTOR,
                    #     up_vector * min_food_offset,
                    #     markers,
                    # )

                    # fix if we see this again, shouldn't really happen
                    raise ImpossibleWorldError("Could not boost spawn high enough to see goal. Very odd.")
                    # break
                visibility_height_offset += height_delta_per_step

            # actually boost the spawn
            spawn_radius = 2.0
            spawn_point_region = world.map.get_dist_sq_to(to_2d_point(solution.locations.spawn)) < spawn_radius**2
            if visibility_height_offset > 0.0:
                map_new = world.map.copy()
                map_new.raise_island(spawn_point_region, visibility_height_offset)
                world = attr.evolve(world, map=map_new)

    # figure out how much the height of the initial building changed, and update its (and its items0 positions
    if initial_building is not None:
        assert initial_building_items is not None
        assert initial_building_starting_terrain_height is not None
        new_height = world.map.get_rough_height_at_point(
            np.array([initial_building.position[0], initial_building.position[2]])
        )
        height_change = new_height - initial_building_starting_terrain_height
        items = move_all_items_up(initial_building_items, height_change)
        world = world.add_items(items)
        new_goal_location = starting_locations.goal + UP_VECTOR * height_change
        world = world.add_item(CANONICAL_FOOD_CLASS(position=new_goal_location))
        assert len(world.building_by_id) == 1
        new_building_position = initial_building.position + np.array([0, height_change, 0])
        new_building = attr.evolve(initial_building, position=new_building_position)
        world = attr.evolve(world, buildings=tuple([new_building]))

    if is_debug_graph_printing_enabled:  # TODO KJ
        logger.debug(f"{obstacle_count=}")
        logger.debug(f"{desired_obstacle_count=}")

    # add any required buildings
    for i in range(building_count):
        # randomly gets a task and location mask for where it can actually fit
        building_obstacle = get_random_building_obstacle(rand, world, task_config, solution.locations, difficulty)
        if building_obstacle:
            task, radius, mask, new_difficulty = building_obstacle
            building_location_2d = world.map.index_to_point_2d(tuple(rand.choice(np.argwhere(mask))))  # type: ignore[arg-type]
            food_location, _building, items, world = create_building(
                rand, world, task_config, building_location_2d, radius, task, difficulty, aesthetics
            )
            world = world.add_items(items)
            world = world.add_item(CANONICAL_FOOD_CLASS(position=food_location))
            obstacle_count += 1

    # finally, add a bunch of predators
    world = add_extra_predators(rand, world, task_config, solution, difficulty, random_predator_count, full_path)

    return _return_data(
        rand, world, difficulty, solution.locations, is_food_required, is_visibility_required, task_config
    )


def add_extra_predators(
    rand: np.random.Generator,
    world: World,
    task_config: CompositeTaskConfig,
    solution: CompositionalSolutionData,
    difficulty: float,
    random_predator_count: int,
    full_path: CompositionalSolutionPath,
) -> World:
    if random_predator_count > 0:

        weapon_ratio = round(
            difficulty_variation(
                task_config.weapon_ratio_easy,
                task_config.weapon_ratio_hard,
                rand,
                difficulty,
            )
        )
        weapon_count = round(weapon_ratio * random_predator_count)

        # probably will never need more than this resolution for path distances
        max_points = 1000
        sq_path_distances = (
            world.map.distances_from_points(
                full_path.point_indices, solution.locations.island, "solution path distances", max_points, rand
            )
            ** 2
        )

        max_distance = scale_with_difficulty(
            difficulty, task_config.predator_path_dist_easy, task_config.predator_path_dist_hard
        )
        max_sq_dist = max_distance**2
        for i in range(random_predator_count):
            position = world.get_safe_point(
                rand, sq_distances=sq_path_distances, max_sq_dist=max_sq_dist, island_mask=solution.locations.island
            )
            if position is not None:
                predator, output_difficulty = get_random_predator(rand, difficulty)
                predator = attr.evolve(predator, position=position)
                world = world.add_item(predator, reset_height_offset=predator.get_offset())

        max_distance = scale_with_difficulty(
            difficulty, task_config.weapon_path_dist_easy, task_config.weapon_path_dist_hard
        )
        max_sq_dist = max_distance**2
        weapon_classes: List[Type[Weapon]] = []
        for i in range(weapon_count):
            safe_mask = world.get_safe_mask(
                sq_distances=sq_path_distances, max_sq_dist=max_sq_dist, island_mask=solution.locations.island
            )
            position = world._get_safe_point(rand, safe_mask)
            rock_probability = scale_with_difficulty(
                difficulty, task_config.rock_probability_easy, task_config.rock_probability_hard
            )
            if rand.uniform() < rock_probability:
                if rand.uniform() < task_config.large_rock_probability:
                    weapon_classes.append(LargeRock)
                else:
                    weapon_classes.append(Rock)
            else:
                weapon_classes.append(Stick)
            if position is not None:
                tool = Placeholder(position=position)
                world = world.add_item(tool, reset_height_offset=tool.get_offset())
        world = world.replace_weapon_placeholders(weapon_classes, solution.locations.island, ITEM_FLATTEN_RADIUS)

    return world


def move_all_items_up(items: List[Entity], height_delta: float) -> List[Entity]:
    new_items = []
    for item in items:
        new_item_position = item.position.copy()
        new_item_position[1] += height_delta
        item = attr.evolve(item, position=new_item_position)
        new_items.append(item)
    return new_items


def remove_close_predators(world: World, spawn: Point2DNP, radius: float) -> World:
    assert_isinstance(spawn, Point2DNP)
    new_items = []
    for item in world.items:
        too_close = False
        if isinstance(item, Predator):
            if np.linalg.norm(to_2d_point(item.position) - spawn) < radius:
                too_close = True
        if not too_close:
            new_items.append(item)
    return attr.evolve(world, items=tuple(new_items))


def get_building_task_generator_class(building_task: BuildingTask) -> Type[BuildingTaskGenerator]:
    generator_by_task = {
        BuildingTask.EXPLORE: ExploreTaskGenerator,
        BuildingTask.OPEN: OpenTaskGenerator,
        BuildingTask.PUSH: PushTaskGenerator,
        BuildingTask.STACK: StackTaskGenerator,
        BuildingTask.CLIMB: ClimbTaskGenerator,
        BuildingTask.JUMP: JumpTaskGenerator,
    }
    if building_task not in generator_by_task:
        raise ValueError(f"No building task generator for {building_task}")
    return generator_by_task[building_task]


def create_building(
    rand: np.random.Generator,
    world: World,
    task_config: CompositeTaskConfig,
    point: Point2DNP,
    radius: float,
    building_task: BuildingTask,
    difficulty: float,
    aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig(),
    is_shore_created: bool = False,
):
    building_location = np.array([point[0], world.map.get_rough_height_at_point(point), point[1]])
    building_task_config = task_config.indoor_obstacle_configs[building_task][1]
    task_generator = get_building_task_generator_class(building_task)(building_task_config)
    building, items, _spawn_location, target_location = create_building_for_skill_scenario(
        rand,
        difficulty,
        task_generator,
        building_location,
        allowed_auxiliary_tasks=(BuildingTask.OPEN, BuildingTask.CLIMB),
        allowed_entrance_tasks=(BuildingTask.OPEN,),
        aesthetics_config=aesthetics,
        is_indoor_only=False,
    )

    # create a flat area around the building
    map_new = world.map.copy()
    map_new.radial_flatten(point, radius * 2)
    # plot_value_grid(map_new.Z, "After flattening for building")

    # also make space for the basement:
    within_radius_mask = map_new.get_dist_sq_to(point) < radius * radius
    points = map_new.get_2d_points()
    within_radius_points = points[within_radius_mask]

    # figure out where the building is in the grid
    building_mask = np.zeros_like(within_radius_mask)
    polygon = np.array([x for x in get_building_footprint_outline(building).exterior.coords])
    building_mask[within_radius_mask] = points_in_polygon(polygon, within_radius_points)

    # if requested, ensure all of those points are on land (other calls must guarantee this)
    if is_shore_created:
        map_new.Z[building_mask] = np.clip(map_new.Z[building_mask], WATER_LINE + 0.1, None)

    # set points near the building outline as special
    building_wall_cell_thickness = round(map_new.cells_per_meter * 2.0)
    building_outline_mask = map_new.get_outline(building_mask, building_wall_cell_thickness)
    is_detail_important_new = world.is_detail_important.copy()
    is_detail_important_new[building_outline_mask] = True
    # plot_value_grid(world.is_detail_important)

    # set height inside the building outline lower depending on the basement height
    basement_mask = map_new.shrink_region(building_mask, building_wall_cell_thickness // 2)
    basement_height = building.stories[0].floor_negative_depth + DEFAULT_FLOOR_THICKNESS + 0.1
    map_new.Z[basement_mask] -= basement_height
    # plot_value_grid(basement_mask)

    # if it is underwater, have to move everything up:
    building_min_height = map_new.Z[basement_mask].min()
    if building_min_height <= WATER_LINE:
        height_delta = -1.0 * building_min_height + 0.1
        map_new.Z[basement_mask] += height_delta
        items = move_all_items_up(items, height_delta)
        new_building_position = building.position + np.array([0, height_delta, 0])
        building = attr.evolve(building, position=new_building_position)

    # prevent trees and stuff from growing into buildings
    world = world.mask_flora(1.0 - map_new.create_radial_mask(point, radius * 2))
    world = attr.evolve(world, map=map_new, is_detail_important=is_detail_important_new)

    building, world = world.add_building(building, basement_mask)

    return target_location, building, items, world


def points_in_polygon(polygon: np.ndarray, pts: np.ndarray) -> np.ndarray:
    pts = np.asarray(pts, dtype="float32")
    polygon = np.asarray(polygon, dtype="float32")
    contour2 = np.vstack((polygon[1:], polygon[:1]))
    test_diff = contour2 - polygon
    mask1 = (pts[:, None] == polygon).all(-1).any(-1)
    m1 = (polygon[:, 1] > pts[:, None, 1]) != (contour2[:, 1] > pts[:, None, 1])
    slope = ((pts[:, None, 0] - polygon[:, 0]) * test_diff[:, 1]) - (
        test_diff[:, 0] * (pts[:, None, 1] - polygon[:, 1])
    )
    m2 = slope == 0
    mask2 = (m1 & m2).any(-1)
    m3 = (slope < 0) != (contour2[:, 1] < polygon[:, 1])
    m4 = m1 & m3
    count = np.count_nonzero(m4, axis=-1)
    mask3 = ~(count % 2 == 0)
    mask = mask1 | mask2 | mask3
    return cast(np.ndarray, mask)


def get_acceptable_building_placement_locations(world: World, safe_mask: MapBoolNP, radius: float) -> MapBoolNP:
    unsafe = np.logical_not(safe_mask)
    radius_cell_dist = int(round(world.map.cells_per_meter * radius)) + 1
    return np.logical_and(safe_mask, np.logical_not(morphology.dilation(unsafe, morphology.disk(radius_cell_dist))))


def get_random_building_obstacle(
    rand: np.random.Generator,
    world: World,
    task_config: CompositeTaskConfig,
    locations: WorldLocations,
    difficulty: float,
) -> Optional[Tuple[BuildingTask, float, MapBoolNP, float]]:
    safe_mask = world.get_safe_mask(locations.island)
    for i in range(_INDOOR_OBSTACLE_MAX_RETRIES):
        task = _select_building_obstacle_type(rand, task_config)
        radius = get_radius_for_building_task(rand, task, difficulty, task_config)
        mask = get_acceptable_building_placement_locations(world, safe_mask, radius)
        if np.any(mask):
            return task, radius, mask, difficulty

    # fine, try one last time with the simplest building possible:
    task = BuildingTask.EXPLORE
    radius = MIN_NAVIGATE_BUILDING_SIZE
    mask = get_acceptable_building_placement_locations(world, safe_mask, radius)
    if np.any(mask):
        return task, radius, mask, difficulty

    return None


def _return_data(
    rand: np.random.Generator,
    world: World,
    difficulty: float,
    locations: WorldLocations,
    is_food_required: bool,
    is_visibility_required: bool,
    task_config: CompositeTaskConfig,
) -> Tuple[World, WorldLocations]:
    if is_food_required:
        world = add_food_tree_for_simple_task(world, locations)
    world = world.add_spawn(
        rand, difficulty, locations.spawn, locations.goal, is_visibility_required=is_visibility_required
    )
    world = remove_close_predators(world, to_2d_point(locations.spawn), task_config.min_predator_distance_from_spawn)
    return world, locations


def _add_height_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    world: World,
    obstacle_type: AvalonTask,
    task_config: CompositeTaskConfig,
    solution: CompositionalSolutionData,
    radius_distance_to_use: float,
    is_around_spawn: bool,
    # TODO: looks like this might just be the negative of the above
    is_height_inverted: bool,
    is_debug_graph_printing_enabled: bool,
) -> Tuple[int, float, CompositionalSolutionData, WorldLocations, World]:
    errors = []

    radius_left = solution.locations.get_2d_spawn_goal_distance() - (
        solution.current_spawn_side_radius + solution.current_goal_side_radius
    )
    if radius_left <= 0.0:
        raise ImpossibleWorldError("Not enough radius left for another height obstacle")
    max_radius_distance_to_use = min([radius_left * 0.99, radius_distance_to_use])
    min_radius_distance_to_use = max_radius_distance_to_use * 0.8

    for i in range(_OUTDOOR_OBSTACLE_MAX_RETRIES):
        radius_step = rand.uniform(min_radius_distance_to_use, max_radius_distance_to_use)
        try:
            if is_around_spawn:
                center = solution.locations.spawn
                start_point = solution.spawn_endpoint
                final_radius = solution.current_spawn_side_radius + radius_step
                end_point = solution.full_path.get_point_at_radius_from_spawn(final_radius)
            else:
                center = solution.locations.goal
                start_point = solution.goal_endpoint
                final_radius = solution.current_goal_side_radius + radius_step
                end_point = solution.full_path.get_point_at_radius_from_goal(final_radius)

            constraint = CompositionalConstraint(
                attr.evolve(solution.locations, spawn=start_point, goal=end_point),
                world,
                center,
                solution.full_path.mask,
                is_height_inverted,
            )

            # if is_debug_graph_printing_enabled:
            #     interesting_points = [constraint.center, constraint.locations.spawn, constraint.locations.goal]
            #     markers = [world.map.point_to_index(np.array([x[0], x[2]])) for x in interesting_points]
            #     plot_terrain(world.map.Z, f"Solution constraint for {obstacle_type}", markers=markers)

            # remember the heights of the goal and spawn before we change stuff
            goal_height_start = world.map.get_rough_height_at_point(to_2d_point(solution.locations.goal))
            spawn_height_start = world.map.get_rough_height_at_point(to_2d_point(solution.locations.spawn))

            inner_task_config = task_config.outdoor_obstacle_configs[obstacle_type][1]

            export_config = world.export_config
            if obstacle_type == AvalonTask.MOVE:
                assert isinstance(inner_task_config, MoveTaskConfig), f"Unexpected type: {type(inner_task_config)}"
                world, new_locations, difficulty = create_move_obstacle(
                    rand, difficulty, export_config, constraint, task_config=inner_task_config
                )
            elif obstacle_type == AvalonTask.JUMP:
                assert isinstance(
                    inner_task_config, JumpOutdoorTaskConfig
                ), f"Unexpected type: {type(inner_task_config)}"
                world, new_locations, difficulty = create_jump_obstacle(
                    rand, difficulty, export_config, constraint, task_config=inner_task_config
                )
            elif obstacle_type == AvalonTask.CLIMB:
                assert isinstance(
                    inner_task_config, ClimbOutdoorTaskConfig
                ), f"Unexpected type: {type(inner_task_config)}"
                world, new_locations, difficulty = create_climb_obstacle(
                    rand, difficulty, export_config, constraint, task_config=inner_task_config
                )
            elif obstacle_type == AvalonTask.PUSH:
                assert isinstance(
                    inner_task_config, PushOutdoorTaskConfig
                ), f"Unexpected type: {type(inner_task_config)}"
                world, new_locations, difficulty = create_push_obstacle(
                    rand, difficulty, export_config, constraint, task_config=inner_task_config
                )
            elif obstacle_type == AvalonTask.FIGHT:
                assert isinstance(inner_task_config, FightTaskConfig), f"Unexpected type: {type(inner_task_config)}"
                world, new_locations, difficulty = create_fight_obstacle(
                    rand, difficulty, export_config, constraint, task_config=inner_task_config
                )
            elif obstacle_type == AvalonTask.STACK:
                assert isinstance(
                    inner_task_config, StackOutdoorTaskConfig
                ), f"Unexpected type: {type(inner_task_config)}"

                world, new_locations, difficulty = create_stack_obstacle(
                    rand, difficulty, export_config, constraint, task_config=inner_task_config
                )
            elif obstacle_type == AvalonTask.BRIDGE:
                assert isinstance(inner_task_config, BridgeTaskConfig), f"Unexpected type: {type(inner_task_config)}"
                world, new_locations, difficulty = create_bridge_obstacle(
                    rand, difficulty, export_config, constraint, task_config=inner_task_config
                )
            elif obstacle_type == AvalonTask.DESCEND:
                assert isinstance(inner_task_config, DescendTaskConfig), f"Unexpected type: {type(inner_task_config)}"
                world, new_locations, difficulty = create_descend_obstacle(
                    rand, difficulty, export_config, constraint, task_config=inner_task_config
                )
            else:
                raise SwitchError(f"Unhandled obstacle type: {obstacle_type}")

            break
        except ImpossibleWorldError as e:
            if i >= _OUTDOOR_OBSTACLE_MAX_RETRIES - 1:
                raise
            errors.append(e)

    # fix up the heights
    goal_height_end = world.map.get_rough_height_at_point(to_2d_point(solution.locations.goal))
    spawn_height_end = world.map.get_rough_height_at_point(to_2d_point(solution.locations.spawn))
    goal_fixed = solution.locations.goal.copy()
    goal_fixed[1] += goal_height_end - goal_height_start
    spawn_fixed = solution.locations.spawn.copy()
    spawn_fixed[1] += spawn_height_end - spawn_height_start
    new_solution = attr.evolve(solution, locations=attr.evolve(solution.locations, spawn=spawn_fixed, goal=goal_fixed))

    # logger.debug(f"Goal height changed by {goal_height_end - goal_height_start}")

    if is_debug_graph_printing_enabled:
        interesting_points = [constraint.center, constraint.locations.spawn, constraint.locations.goal]
        markers = [world.map.point_to_index(np.array([x[0], x[2]])) for x in interesting_points]
        plot_terrain(world.map.Z, f"Solution constraint for {obstacle_type}", markers=markers)

    # new solution needs to figure out  max point along the path given the new radius
    if is_around_spawn:
        new_solution = new_solution.update_spawn(final_radius)
    else:
        new_solution = new_solution.update_goal(final_radius)
    return 1, difficulty, new_solution, new_locations, world


def _find_path(rand: np.random.Generator, world: World, locations: WorldLocations) -> CompositionalSolutionPath:
    start_2d = to_2d_point(locations.spawn)
    end_2d = to_2d_point(locations.goal)
    path = _find_path_from_points(rand, world, start_2d, end_2d)
    mask = np.zeros_like(world.is_climbable)
    point_indices = np.array(path)
    sel = point_indices[:, 0], point_indices[:, 1]
    mask[sel] = 1
    points_2d = np.stack([world.map.X[sel], world.map.Y[sel]], axis=1)
    points = np.stack([world.map.X[sel], world.map.Z[sel], world.map.Y[sel]], axis=1)
    start_2d_radii = np.linalg.norm(points_2d - start_2d, axis=-1)
    end_2d_radii = np.linalg.norm(points_2d - end_2d, axis=-1)
    path_locations = np.linspace(0.0, 1.0, num=len(end_2d_radii))
    return CompositionalSolutionPath(
        mask=mask,
        path_locations=path_locations,
        points=points,
        point_indices=point_indices,
        start_2d_radii=start_2d_radii,
        end_2d_radii=end_2d_radii,
    )


def _find_path_from_points(
    rand: np.random.Generator, world: World, start: Point2DNP, end: Point2DNP
) -> List[Tuple[int, int]]:
    ngrid = len(world.map.Z)
    # graph = nx.grid_graph(dim=[ngrid, ngrid])
    graph = nx.grid_2d_graph(ngrid, ngrid)

    max_water_dist = 20.0
    max_penalty = 100000
    sqrt2 = math.sqrt(2)

    water_distances = world.map.get_water_distance(
        rand,
        is_fresh_water_included_in_moisture=True,
        max_points=world.biome_config.max_kd_points if world.biome_config else None,
    )

    def cost(a: Tuple[int, int], b: Tuple[int, int], k1: float = 1.0, k2: float = 10.0, kind: str = "intsct") -> float:
        dist = 1 if (a[0] == b[0] or a[1] == b[1]) else sqrt2
        assert water_distances is not None
        water_distance = water_distances[b]
        if water_distance <= 0.01:
            penalty = np.inf
        elif water_distance < max_water_dist:
            factor = 1.0 - (water_distance / max_water_dist) ** 3
            penalty = max_penalty * factor
        else:
            penalty = 0.0
        return dist + penalty

    start_indices = world.map.point_to_index(start)
    end_indices = world.map.point_to_index(end)
    return cast(List[Tuple[int, int]], nx.astar_path(graph, start_indices, end_indices, _get_2d_dist, weight=cost))


def _get_2d_dist(a: Tuple[int, int], b: Tuple[int, int]) -> float:
    x1 = np.array(a, dtype=np.float32)
    x2 = np.array(b, dtype=np.float32)
    return cast(float, np.linalg.norm(x1 - x2))


def _select_ring_obstacle_type(
    rand: np.random.Generator,
    allowed_obstacles: Tuple[AvalonTask, ...],
    task_config: CompositeTaskConfig,
) -> AvalonTask:
    obstacles = allowed_obstacles
    probabilities = []
    for obstacle in allowed_obstacles:
        probabilities.append(task_config.outdoor_obstacle_configs[obstacle][0])
    probabilities_np = np.array(probabilities)
    probabilities_np /= probabilities_np.sum()
    return rand.choice(obstacles, p=probabilities_np)  # type: ignore


def _select_building_obstacle_type(
    rand: np.random.Generator,
    task_config: CompositeTaskConfig,
) -> BuildingTask:
    obstacles = [x for x in task_config.indoor_obstacle_configs.keys()]
    probabilities = []
    for obstacle in obstacles:
        probabilities.append(task_config.indoor_obstacle_configs[obstacle][0])
    probabilities_np = np.array(probabilities)
    probabilities_np /= probabilities_np.sum()
    return rand.choice(obstacles, p=probabilities_np)  # type: ignore


def _select_final_obstacle_type(
    rand: np.random.Generator,
    difficulty: float,
    obstacles: List[FinalObstacleType],
    _FORCED: Optional[FinalObstacleType] = None,
) -> FinalObstacleType:
    probabilities = [1.0] * len(obstacles)
    probabilities_np = np.array(probabilities)
    probabilities_np /= probabilities_np.sum()
    return rand.choice(obstacles, p=probabilities_np)  # type: ignore


def _get_desired_obstacle_count(
    rand: np.random.Generator,
    difficulty: float,
    task_config: CompositeTaskConfig,
    _FORCED: Optional[ForcedComposition],
) -> Tuple[int, float]:
    if task_config.task == AvalonTask.SURVIVE:
        if _FORCED and _FORCED.is_enabled:
            desired_obstacle_count = 1 if difficulty > 0.5 else 0
        else:
            is_obstacle_present, difficulty = select_boolean_difficulty(difficulty, rand)
            desired_obstacle_count = 1 if is_obstacle_present else 0
    else:
        if _FORCED and _FORCED.is_enabled:
            desired_obstacle_count = round(scale_with_difficulty(difficulty, 1.5, task_config.max_obstacles + 0.5))
        else:
            desired_obstacle_count = round(
                normal_distrib_range(
                    1.25, task_config.max_obstacles + 0.49, task_config.max_obstacles_std_dev, rand, difficulty
                )
            )
    return desired_obstacle_count, difficulty


def _get_obstacle_counts_per_type(
    rand: np.random.Generator, desired_obstacle_count: int, task_config: CompositeTaskConfig, is_food_allowed: bool
) -> Tuple[int, int]:
    if desired_obstacle_count == 0:
        return 0, 0
    is_food_in_building = rand.uniform() < task_config.is_goal_in_building_probability
    building_count = 1 if is_food_in_building else 0
    if not is_food_allowed:
        building_count = 0
    remaining = desired_obstacle_count - building_count
    # can only create new food (ie, buildings) if this is a gather or survive task
    if task_config.task in (AvalonTask.GATHER, AvalonTask.SURVIVE) and is_food_allowed:
        count_one = rand.integers(0, remaining, endpoint=True)
        count_two = remaining - count_one
        values = np.array([count_one, count_two])
        rand.shuffle(values)
        extra_building_count, height_obstacle_count = list(values)
        building_count += extra_building_count
    else:
        height_obstacle_count = remaining
    return building_count, height_obstacle_count


def _get_fall_climb_split(rand: np.random.Generator, desired_height_obstacle_count: int) -> Tuple[int, int]:
    if desired_height_obstacle_count == 0:
        return 0, 0
    fall_count = rand.integers(0, desired_height_obstacle_count, endpoint=True)
    climb_count = desired_height_obstacle_count - fall_count
    return fall_count, climb_count


def get_random_predator(rand: np.random.Generator, difficulty: float):
    predator_class, new_difficulty = select_categorical_difficulty(ALL_PREDATOR_CLASSES, difficulty, rand)
    predator = predator_class(position=np.array([0.0, 0.0, 0.0]))
    return predator, new_difficulty


def get_radius_for_building_task(
    rand: np.random.Generator, task: BuildingTask, difficulty: float, task_config: Optional[CompositeTaskConfig] = None
) -> float:
    generator_kwargs = {}
    if task_config is not None:
        generator_kwargs["config"] = task_config.indoor_obstacle_configs[task][1]
    generator_class = get_building_task_generator_class(task)
    return generator_class(**generator_kwargs).get_site_radius(rand, difficulty)



from typing import Type

from avalon.agent.common.params import Params
from avalon.agent.common.types import Algorithm
from avalon.agent.dreamer.params import DreamerParams
from avalon.agent.ppo.params import PPOParams
from avalon.contest.contest_params import ContestAlgorithmParams


def get_algorithm_cls(params: Params) -> Type[Algorithm]:
    """This is a bit hacky because we don't really want to import code from an algorithm we're not using."""
    algorithm_cls: Type[Algorithm]
    if issubclass(type(params), PPOParams):
        from avalon.agent.ppo.ppo import PPO

        algorithm_cls = PPO
    elif issubclass(type(params), DreamerParams):
        from avalon.agent.dreamer.dreamer import Dreamer

        algorithm_cls = Dreamer
    elif issubclass(type(params), ContestAlgorithmParams):
        from avalon.contest.eval import ContestAlgorithmWrapper

        algorithm_cls = ContestAlgorithmWrapper
    else:
        raise NotImplementedError(type(params).__name__)
    return algorithm_cls



"""This is an alternative implementation of the rollout worker.
Used to verify the output of the primary rollout worker.
Intended to be simple enough to be able to readily verify correctness,
but still support all the features we want to test.

The main simplification is just in not handling partial batches.
It's also lacks the complexity from performance tuning.
"""
from collections import defaultdict
from multiprocessing.context import BaseContext
from typing import Optional
from typing import Type

import attr
import gym
import torch
from torch import Tensor

from avalon.agent.common.envs import build_env
from avalon.agent.common.params import Params
from avalon.agent.common.storage import TrajectoryStorage
from avalon.agent.common.types import Algorithm
from avalon.agent.common.types import StepData
from avalon.agent.common.util import postprocess_uint8_to_float


class DummyStorage(TrajectoryStorage):
    def __init__(self, params: Params, step_data_type: Type[StepData], num_workers: int) -> None:
        super().__init__(params, step_data_type, num_workers)
        self.storage = None

    def to_packed(self):
        assert self.storage is not None
        return self.step_data_type.batch_sequence_type(**self.storage)

    def reset(self) -> None:
        self.storage = None


class SimpleRolloutManager:
    def __init__(
        self,
        params: Params,
        num_workers: int,
        is_multiprocessing: bool,
        storage: TrajectoryStorage,
        obs_space: gym.spaces.Dict,
        model: Algorithm,
        rollout_device: torch.device,
        multiprocessing_context: BaseContext,
    ) -> None:
        # TODO: handle the context / fork/spawn method somehow
        vector_env_cls = gym.vector.AsyncVectorEnv if is_multiprocessing else gym.vector.SyncVectorEnv
        self.num_workers = num_workers
        self.params = params
        self.model = model
        self.rollout_device = rollout_device
        self.storage = storage

        params = attr.evolve(params, env_params=attr.evolve(params.env_params, env_count=num_workers))
        # the torchify wrapper doesn't work, they still get coerced to numpy in the VecEnv
        self.envs = vector_env_cls(
            [
                lambda i=i: build_env(
                    env_params=attr.evolve(params.env_params, env_index=params.env_params.env_index + i),
                    torchify=False,
                )
                for i in range(num_workers)
            ]
        )

        # We need to carry these between rollout batches
        obs_numpy = self.envs.reset()
        self.obs = {k: torch.from_numpy(v) * 0 for k, v in obs_numpy.items()}
        self.dones = torch.zeros((self.num_workers,), dtype=torch.bool, device=self.rollout_device)

    def run_rollout(
        self,
        num_steps: Optional[int] = None,
        num_episodes: Optional[int] = None,
        exploration_mode: str = "explore",
    ) -> None:
        storage = defaultdict(list)
        storage["observation"] = defaultdict(list)
        storage["action"] = defaultdict(list)

        assert num_episodes is None
        self.model.eval()
        to_run = torch.ones((self.num_workers,), dtype=torch.bool)
        for i in range(num_steps):
            with torch.no_grad():
                torch_obs = {k: v.to(device=self.rollout_device, non_blocking=True) for k, v in self.obs.items()}
                torch_dones = self.dones.to(device=self.rollout_device, non_blocking=True)
                torch_obs = postprocess_uint8_to_float(torch_obs, center=self.params.center_observations)
                actions, to_store = self.model.rollout_step(
                    torch_obs, torch_dones, to_run, exploration_mode=exploration_mode
                )

            assert self.params.obs_first is True
            for k, v in self.obs.items():
                storage["observation"][k].append(v)
            for k, v in actions.items():
                storage["action"][k].append(v)
            actions_numpy = {k: v.numpy() for k, v in actions.items()}
            obs_numpy, rewards, dones, infos = self.envs.step(actions_numpy)
            self.obs = {k: torch.from_numpy(v) for k, v in obs_numpy.items()}
            rewards = torch.from_numpy(rewards).to(dtype=torch.float32)
            dones = torch.from_numpy(dones).to(dtype=torch.bool)
            self.dones = dones
            is_terminal = dones.clone()

            if self.params.time_limit_bootstrapping:
                # infos is a tuple of info dicts
                for worker_id, done in enumerate(dones):
                    if "TimeLimit.truncated" in infos:
                        # Note: i think i'm playing a little loose with how the infos work here.
                        # They also have a _field_name bool parameter that probably says something like
                        # if that worker_id gave an info?
                        is_terminal[worker_id] = done and not infos["TimeLimit.truncated"][worker_id]
                        if done and not is_terminal[worker_id]:
                            # terminal_observation field name changed to final_observation
                            # in a gym update; seems in the version we're currently using the vector env
                            # got the new name while regular envs still have the old name.
                            terminal_observation = infos["final_observation"][worker_id]
                            additional_reward = self.time_limit_bootstrapping(terminal_observation)
                            rewards[worker_id] = rewards[worker_id] + additional_reward

            storage["reward"].append(rewards)
            storage["done"].append(dones)
            storage["is_terminal"].append(is_terminal)
            for k, v in to_store.items():
                storage[k].append(v)

        # we include one extra observation
        for k, v in self.obs.items():
            storage["observation"][k].append(v)

        # Pack everything into a single tensor per (nested) key
        for k, v in storage.items():
            if k in ("observation", "action"):
                storage[k] = {k: torch.stack(v, dim=1) for k, v in storage[k].items()}
            else:
                storage[k] = torch.stack(v, dim=1)

        self.storage.storage = storage

    def time_limit_bootstrapping(self, terminal_obs: dict[str, Tensor]) -> float:
        # NOTE: this won't work (as written) for recurrent algorithms!!
        # value bootstrapping for envs with artificial time limits imposed by a gym TimeLimit wrapper
        # see https://arxiv.org/pdf/1712.00378.pdf for context
        # this does get wrapper obs transforms applied
        terminal_obs_torch = {
            k: torch.from_numpy(v).to(self.rollout_device).unsqueeze(0) for k, v in terminal_obs.items()
        }
        terminal_obs_torch = postprocess_uint8_to_float(terminal_obs_torch, center=self.params.center_observations)
        with torch.no_grad():
            terminal_value, _ = self.model(terminal_obs_torch)
        return self.params.discount * terminal_value[0].detach().cpu().item()  # type: ignore

    def shutdown(self) -> None:
        try:
            self.envs.close()
        except BrokenPipeError:
            pass



from typing import NamedTuple
from typing import Tuple
from typing import cast

import numpy as np

from avalon.datagen.world_creation.indoor.constants import WallType
from avalon.datagen.world_creation.types import Point3DNP


class WallFootprint(NamedTuple):
    top_left: Tuple[int, int]  # x, z
    bottom_right: Tuple[int, int]  # x, z
    wall_type: WallType
    is_vertical: bool

    @property
    def wall_thickness(self) -> int:
        return self.footprint_width if self.is_vertical else self.footprint_length

    @property
    def wall_length(self) -> int:
        return self.footprint_length if self.is_vertical else self.footprint_width

    @property
    def footprint_width(self) -> int:
        return self.bottom_right[0] - self.top_left[0]

    @property
    def footprint_length(self) -> int:
        return self.bottom_right[1] - self.top_left[1]

    @property
    def centroid(self) -> Point3DNP:
        return cast(Point3DNP, np.array(self.top_left) + (np.array(self.bottom_right) - np.array(self.top_left)) / 2)



import json
from hashlib import md5
from numbers import Number
from typing import Generic
from typing import Tuple
from typing import TypeVar

from avalon.common.utils import find_root_dir
from avalon.common.utils import hash_dirpath
from avalon.contrib.serialization import Serializable

RangeT = TypeVar("RangeT", bound=Number)


class AbstractRange(Generic[RangeT], Serializable):
    min_ge: RangeT
    max_lt: RangeT

    @property
    def size(self) -> RangeT:
        return self.max_lt - self.min_ge  # type: ignore


DataT = TypeVar("DataT", bound=Serializable)


class AbstractDataConfig(Serializable):
    def get_is_using_shared_caches(self) -> bool:
        raise NotImplementedError()

    def set_is_using_shared_caches(self: DataT, is_using_shared_caches: bool) -> DataT:
        raise NotImplementedError()

    def get_random_key(self) -> str:
        raise NotImplementedError()

    def get_dir_root(self) -> str:
        raise NotImplementedError()

    def set_dir_root(self: DataT, dir_root: str) -> DataT:
        raise NotImplementedError()

    def get_video_range(self) -> AbstractRange:
        raise NotImplementedError()

    def get_features(self) -> Tuple[str, ...]:
        raise NotImplementedError()

    def get_frame_range(self) -> AbstractRange:
        raise NotImplementedError()

    def create_data_generation_request(
        self: DataT, video_range: AbstractRange, random_key: str, dir_root: str
    ) -> DataT:
        raise NotImplementedError()

    def get_output_file_name_format(self) -> str:
        raise NotImplementedError()

    def get_is_generating_paired_videos(self) -> bool:
        raise NotImplementedError()


def get_datagen_hash(config: Serializable) -> str:
    # TODO: make sure there's no extra usages, since the Godot binary path is different for old datagen
    hashed_config = md5(json.dumps(config.to_dict(), sort_keys=True).encode("utf-8")).hexdigest()
    hashed_godot_binary = md5(open("/usr/bin/godot-old", "rb").read()).hexdigest()
    hashed_godot_dir = hash_dirpath(find_root_dir() + "/datagen/godot/**/*")
    return md5((hashed_godot_dir + hashed_config + hashed_godot_binary).encode("utf-8")).hexdigest()



from io import BufferedReader
from typing import Callable
from typing import Generic
from typing import Iterator
from typing import List
from typing import Literal
from typing import Set
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

import attr
from loguru import logger

from avalon.common.errors import SwitchError
from avalon.datagen.godot_env.actions import ActionType
from avalon.datagen.godot_env.actions import DebugCameraAction
from avalon.datagen.godot_env.actions import _from_bytes
from avalon.datagen.godot_generated_types import ACTION_MESSAGE
from avalon.datagen.godot_generated_types import CLOSE_MESSAGE
from avalon.datagen.godot_generated_types import DEBUG_CAMERA_ACTION_MESSAGE
from avalon.datagen.godot_generated_types import HUMAN_INPUT_MESSAGE
from avalon.datagen.godot_generated_types import LOAD_SNAPSHOT_MESSAGE
from avalon.datagen.godot_generated_types import QUERY_AVAILABLE_FEATURES_MESSAGE
from avalon.datagen.godot_generated_types import RENDER_MESSAGE
from avalon.datagen.godot_generated_types import RESET_MESSAGE
from avalon.datagen.godot_generated_types import SAVE_SNAPSHOT_MESSAGE
from avalon.datagen.godot_generated_types import SEED_MESSAGE
from avalon.datagen.godot_generated_types import SELECT_FEATURES_MESSAGE

_BINARY_READ: Literal["br"] = "br"

_NoPayloadMessageTypes = Literal[2, 5, 6, 10]
_NoPayloadMessage = Tuple[_NoPayloadMessageTypes]
_SeedMessage = Tuple[Literal[1], int]
_SelectFeaturesMessage = Tuple[Literal[4], List[str]]
_LoadSnapshotMessage = Tuple[Literal[11], str]
_ActionMessage = Tuple[Literal[3, 9], ActionType]
_DebugMessage = Tuple[Literal[7], DebugCameraAction]
_ResetMessage = Tuple[Literal[3, 0], ActionType, int, str, float]
_RawMessage = Union[
    _NoPayloadMessage,
    _SeedMessage,
    _SelectFeaturesMessage,
    _ActionMessage[ActionType],
    _ResetMessage[ActionType],
    _DebugMessage,
    _LoadSnapshotMessage,
]
_no_payload_messages: Set[_NoPayloadMessageTypes] = {
    RENDER_MESSAGE,
    QUERY_AVAILABLE_FEATURES_MESSAGE,
    CLOSE_MESSAGE,
    SAVE_SNAPSHOT_MESSAGE,
}

_ActionOrDebugMessage = Union[_ActionMessage[ActionType], _DebugMessage]


def _parse_raw_message_log(
    record_log: BufferedReader,
    action_from_bytes: Callable[[bytes], ActionType],
) -> Iterator[_RawMessage[ActionType]]:
    while message_bytes := record_log.read(1):
        message = int.from_bytes(message_bytes, byteorder="little", signed=False)
        if message in _no_payload_messages:
            # TODO remove cast when mypy can refine types (https://github.com/python/mypy/issues/12535)
            yield cast(_NoPayloadMessage, (message,))

        elif message == SEED_MESSAGE:
            episode_seed = int.from_bytes(record_log.read(8), byteorder="little", signed=True)
            yield cast(_SeedMessage, (message, episode_seed))

        elif message == SELECT_FEATURES_MESSAGE:
            count = int.from_bytes(record_log.read(4), byteorder="little", signed=False)
            feature_names = list(record_log.readline().decode("UTF-8")[:-1] for _ in range(count))
            yield cast(_SelectFeaturesMessage, (message, feature_names))

        elif message in (
            ACTION_MESSAGE,
            HUMAN_INPUT_MESSAGE,
            DEBUG_CAMERA_ACTION_MESSAGE,
        ):
            size_bytes = record_log.read(4)
            size, _ = _from_bytes(int, size_bytes)
            action_bytes = size_bytes + record_log.read(cast(int, size))
            yield cast(_ActionOrDebugMessage, (message, action_from_bytes(action_bytes)))

        elif message == RESET_MESSAGE:
            size_bytes = record_log.read(4)
            size, _ = _from_bytes(int, size_bytes)
            action_bytes = size_bytes + record_log.read(cast(int, size))
            episode_seed = int.from_bytes(record_log.read(8), byteorder="little", signed=True)
            world_path = record_log.readline()
            starting_hit_points_bytes = record_log.read(4)
            starting_hit_points = _from_bytes(float, starting_hit_points_bytes)
            yield cast(
                _ResetMessage,
                (message, action_from_bytes(action_bytes), episode_seed, world_path, starting_hit_points),
            )
        elif message == LOAD_SNAPSHOT_MESSAGE:
            snapshot_path = record_log.readline()
            yield cast(_LoadSnapshotMessage, (message, snapshot_path))
        else:
            raise SwitchError(f"Invalid message type {message}")


@attr.s(auto_attribs=True, frozen=True, slots=True)
class GodotEnvActionLog(Generic[ActionType]):
    path: str
    selected_features: List[str]
    initial_episode_id: int
    messages: List[_RawMessage[ActionType]]

    @classmethod
    def parse(cls, action_log_path: str, action_type: Type[ActionType]) -> "GodotEnvActionLog[ActionType]":
        avaliable_features_query, select_features, initial_seed, *messages, close = cls.parse_message_log(
            action_log_path, action_type
        )
        assert avaliable_features_query[0] == QUERY_AVAILABLE_FEATURES_MESSAGE, (
            f"First logged message was {avaliable_features_query} but should always be a "
            f"QUERY_AVAILABLE_FEATURES_MESSAGE ({QUERY_AVAILABLE_FEATURES_MESSAGE})"
        )
        assert select_features[0] == SELECT_FEATURES_MESSAGE, (
            f"Second logged message was {select_features} but should always be a "
            f"SELECT_FEATURES_MESSAGE ({SELECT_FEATURES_MESSAGE})"
        )
        assert initial_seed[0] == SEED_MESSAGE, (
            f"Third logged message was {initial_seed} but should always be a "
            f"SELECT_FEATURES_MESSAGE ({SEED_MESSAGE})"
        )
        if close[0] != CLOSE_MESSAGE:
            logger.warning(f"{action_log_path}'s last logged message {close} is not a CLOSE_MESSAGE")
            messages.append(close)
        return cls(action_log_path, select_features[1], initial_seed[1], messages)

    @classmethod
    def parse_message_log(cls, action_log_path: str, action_type: Type[ActionType]) -> List[_RawMessage[ActionType]]:
        action_parser: Callable[[bytes], ActionType] = action_type.from_bytes
        with open(action_log_path, _BINARY_READ) as log:
            return list(_parse_raw_message_log(log, action_parser))



import attr
import numpy as np


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class FloraConfig:
    # which 3D file to use for the geometry
    resource: str
    # whether or not to use the noise shader with perlin noise
    is_noise_shader_enabled: bool = False
    # whether this scenery should cast shadows
    is_shadowed: bool = True
    # the default scale for this 3D model
    default_scale: float = 1.0
    # how large the collision box is at the default scale. Only applies to trees
    collision_extents: np.ndarray = np.array([1.0, 5.0, 1.0])
    # how much to move the object in the vertical access when placing.
    # helps ensures that things are not floating off of highly sloped terrain
    # only applies to trees
    height_offset: float = 0.0



import attr

from avalon.datagen.world_creation.entities.tools.tool import Tool


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Log(Tool):
    mass: float = 60.0
    resource_file: str = "res://items/log.tscn"

    def get_offset(self) -> float:
        return 0.7  # need to scale this if we scale size
